"""
Reverse everything you say (``Je proteste énergiquement`` will become
``tnemeuqigrené etsetorp eJ``)
"""
from poezio.plugin import BasePlugin
from poezio import xhtml
from slixmpp import Message
from typing import Any


class Plugin(BasePlugin):
    def init(self) -> None:
        self.api.add_event_handler('muc_say', self.revstr)
        self.api.add_event_handler('conversation_say', self.revstr)
        self.api.add_event_handler('private_say', self.revstr)

    def revstr(self, msg: Message, tab: Any) -> None:
        msg['body'] = xhtml.clean_text(msg['body'])[::-1]



"""
This plugin sends a small image to the recipient of your choice, using XHTML-IM and Bits of Binary.

Usage
-----

/bob some/image.png

Configuration options
---------------------

.. glossary::

    max_size
        **Default:** ``2048``

        The maximum acceptable size of a file, over which you will get an error instead.

    max_age
        **Default:** ``86400``

        The time during which the file should stay in cache on the receiving side.
"""

from poezio.core.structs import Completion
from poezio.plugin import BasePlugin
from poezio.windows.inputs import Input
from poezio import tabs

from pathlib import Path
from glob import glob
from os.path import expanduser
from mimetypes import guess_type


class Plugin(BasePlugin):

    default_config = {'bob': {'max_size': 2048, 'max_age': 86400}}

    def init(self) -> None:
        for tab in tabs.DynamicConversationTab, tabs.StaticConversationTab, tabs.PrivateTab, tabs.MucTab:
            self.api.add_tab_command(
                tab,
                'bob',
                self.command_bob,
                usage='<image>',
                help='Send image <image> to the current discussion',
                short='Send a short image',
                completion=self.completion_bob)

    async def command_bob(self, filename: str) -> None:
        path = Path(expanduser(filename))
        try:
            size = path.stat().st_size
        except OSError as exc:
            self.api.information('Error sending “%s”: %s' % (path.name, exc),
                                 'Error')
            return
        mime_type = guess_type(path.as_posix())[0]
        if mime_type is None or not mime_type.startswith('image/'):
            self.api.information(
                'Error sending “%s”, not an image file.' % path.name, 'Error')
            return
        if size > self.config.get('max_size'):
            self.api.information(
                'Error sending “%s”, file too big.' % path.name, 'Error')
            return
        with open(path.as_posix(), 'rb') as file:
            data = file.read()
        max_age = self.config.get('max_age')
        cid = await self.core.xmpp.plugin['xep_0231'].set_bob(
            data, mime_type, max_age=max_age)
        self.api.run_command(
            '/xhtml <img src="cid:%s" alt="%s"/>' % (cid, path.name))

    @staticmethod
    def completion_bob(the_input: Input) -> Completion:
        txt = expanduser(the_input.get_text()[5:])
        images = []
        for filename in glob(txt + '*'):
            mime_type = guess_type(filename)[0]
            if mime_type is not None and mime_type.startswith('image/'):
                images.append(filename)
        return Completion(the_input.auto_completion, images, quotify=False)



"""
Module defining all the "info wins", ie the bar which is on top of the
info buffer in normal tabs
"""

from __future__ import annotations

from collections.abc import Mapping
from typing import Optional, Dict, TYPE_CHECKING, Any, Callable, Protocol

import logging

from slixmpp import JID, InvalidJID

from poezio.config import config

from poezio.windows.base_wins import Win
from poezio.ui.funcs import truncate_nick
from poezio.theming import get_theme, to_curses_attr, TextStyle
from poezio.colors import ccg_text_to_color
from poezio.contact import Contact, Resource

if TYPE_CHECKING:
    from poezio.user import User
    from poezio.tabs import MucTab
    from poezio.windows import TextWin

log = logging.getLogger(__name__)


AdditionalCb = Callable[[str], str]


class HasPos(Protocol):
    """Protocol for a class that has a pos attribute or property"""
    @property
    def pos(self) -> int:
        ...


class InfoWin(Win):
    """
    Base class for all the *InfoWin, used in various tabs. For example
    MucInfoWin, etc. Provides some useful methods.
    """

    __slots__ = ()

    def print_scroll_position(self, window: 'HasPos') -> None:
        """
        Print, like in Weechat, a -MORE(n)- where n
        is the number of available lines to scroll
        down
        """
        if window.pos > 0:
            plus = ' -MORE(%s)-' % window.pos
            self.addstr(plus,
                        to_curses_attr(get_theme().COLOR_SCROLLABLE_NUMBER))


class XMLInfoWin(InfoWin):
    """
    Info about the latest xml filter used and the state of the buffer.
    """

    __slots__ = ()

    def refresh(self, window: 'TextWin', filter_t: str = '', filter: str = '') -> None:
        log.debug('Refresh: %s', self.__class__.__name__)
        self._win.erase()
        bar = to_curses_attr(get_theme().COLOR_INFORMATION_BAR)
        if not filter_t:
            self.addstr('[No filter]', bar)
        else:
            info = '[%s] %s' % (filter_t, filter)
            self.addstr(info, bar)
        self.print_scroll_position(window)
        self.finish_line(get_theme().COLOR_INFORMATION_BAR)
        self._refresh()


class PrivateInfoWin(InfoWin):
    """
    The line above the information window, displaying information
    about the MUC user we are talking to
    """

    __slots__ = ()

    def refresh(self, name: str, window: 'TextWin', chatstate: str | None,
                information: Mapping[str, AdditionalCb]) -> None:
        log.debug('Refresh: %s', self.__class__.__name__)
        self._win.erase()
        self.write_room_name(name)
        self.print_scroll_position(window)
        self.write_chatstate(chatstate)
        self.write_additional_information(information, name)
        self.finish_line(get_theme().COLOR_INFORMATION_BAR)
        self._refresh()

    def write_additional_information(self, information: Mapping[str, AdditionalCb], jid: JID | str) -> None:
        """
        Write all information added by plugins by getting the
        value returned by the callbacks.
        """
        for plugin in information.values():
            self.addstr(plugin(str(jid)),
                        to_curses_attr(get_theme().COLOR_INFORMATION_BAR))

    def write_room_name(self, name: str) -> None:
        # TODO: autocolour this too, but we need more info about the occupant
        # (whether we know its real jid) and the room (whether it is
        # anonymous) to provide correct colouring.
        try:
            jid = JID(name)
        except InvalidJID:
            jid = JID('')
        room_name, nick = jid.bare, jid.resource
        theme = get_theme()
        self.addstr(nick, to_curses_attr(theme.COLOR_PRIVATE_NAME))
        txt = ' from room %s' % room_name
        self.addstr(txt, to_curses_attr(theme.COLOR_INFORMATION_BAR))

    def write_chatstate(self, state: str | None) -> None:
        if state:
            self.addstr(' %s' % (state, ),
                        to_curses_attr(get_theme().COLOR_INFORMATION_BAR))


class MucListInfoWin(InfoWin):
    """
    The live above the information window, displaying informatios
    about the muc server being listed
    """

    __slots__ = ('message')

    def __init__(self, message: str = '') -> None:
        InfoWin.__init__(self)
        self.message = message

    def refresh(self, name: str | None = None, window: HasPos | None = None) -> None:
        log.debug('Refresh: %s', self.__class__.__name__)
        self._win.erase()
        theme = get_theme()
        if name:
            self.addstr(name,
                        to_curses_attr(theme.COLOR_INFORMATION_BAR))
        else:
            self.addstr(self.message,
                        to_curses_attr(theme.COLOR_INFORMATION_BAR))
        if window:
            self.print_scroll_position(window)
        self.finish_line(theme.COLOR_INFORMATION_BAR)
        self._refresh()


class ConversationInfoWin(InfoWin):
    """
    The line above the information window, displaying information
    about the user we are talking to
    """

    __slots__ = ()

    def __init__(self) -> None:
        InfoWin.__init__(self)

    def refresh(self, jid: JID | str, contact: Contact | None,
                window: 'TextWin', chatstate: str | None,
                information: Mapping[str, AdditionalCb]) -> None:
        # contact can be None, if we receive a message
        # from someone not in our roster. In this case, we display
        # only the maximum information from the message we can get.
        log.debug('Refresh: %s', self.__class__.__name__)
        try:
            jid = JID(jid)
        except InvalidJID:
            jid = JID('')
        if contact:
            if jid.resource:
                resource = contact[jid.full]
            else:
                resource = contact.get_highest_priority_resource()
        else:
            resource = None
        # if contact is None, then resource is None too:
        # user is not in the roster so we know almost nothing about it
        # If contact is a Contact, then
        # resource can now be a Resource: user is in the roster and online
        # or resource is None: user is in the roster but offline
        self._win.erase()
        if config.getbool('show_jid_in_conversations'):
            self.write_contact_jid(jid)
        self.write_contact_information(contact)
        self.write_resource_information(resource)
        self.print_scroll_position(window)
        self.write_chatstate(chatstate)
        self.write_additional_information(information, jid)
        self.finish_line(get_theme().COLOR_INFORMATION_BAR)
        self._refresh()

    def write_additional_information(self, information: Mapping[str, AdditionalCb], jid: JID) -> None:
        """
        Write all information added by plugins by getting the
        value returned by the callbacks.
        """
        color = to_curses_attr(get_theme().COLOR_INFORMATION_BAR)
        for plugin in information.values():
            self.addstr(plugin(str(jid)), color)

    def write_resource_information(self, resource: "Resource | None") -> None:
        """
        Write the information about the resource
        """
        if not resource:
            presence = "unavailable"
        else:
            presence = resource.presence
        theme = get_theme()
        color = theme.color_show(presence)
        if not presence:
            presence = theme.CHAR_STATUS
        self.addstr('[', to_curses_attr(theme.COLOR_INFORMATION_BAR))
        self.addstr(presence, to_curses_attr(color))
        if resource and resource.status:
            shortened = resource.status[:20] + (resource.status[:20] and '…')
            self.addstr(' %s' % shortened,
                        to_curses_attr(theme.COLOR_INFORMATION_BAR))
        self.addstr(']', to_curses_attr(theme.COLOR_INFORMATION_BAR))

    def write_contact_information(self, contact: Contact | None) -> None:
        """
        Write the information about the contact
        """
        theme = get_theme()
        color = to_curses_attr(theme.COLOR_INFORMATION_BAR)
        name_color: int | TextStyle
        if config.get('autocolor_tab_names') and contact is not None:
            name_color = (
                ccg_text_to_color(theme.ccg_palette, str(contact.bare_jid)),
                -1,
                theme.MODE_TAB_NAME,
            )
        else:
            name_color = color

        if not contact:
            self.addstr("(contact not in roster)", color)
            return
        display_name = contact.name
        if display_name:
            self.addstr('%s ' % (display_name), name_color)

    def write_contact_jid(self, jid: JID) -> None:
        """
        Just write the jid that we are talking to
        """
        theme = get_theme()
        color: int = to_curses_attr(theme.COLOR_INFORMATION_BAR)
        name_color: TextStyle
        if config.get('autocolor_tab_names'):
            name_color = (
                ccg_text_to_color(theme.ccg_palette, str(jid)),
                -1,
                theme.MODE_TAB_NAME,
            )
        else:
            name_color = theme.COLOR_CONVERSATION_NAME

        self.addstr('[', color)
        self.addstr(jid.full, to_curses_attr(name_color))
        self.addstr('] ', color)

    def write_chatstate(self, state: str | None) -> None:
        if state:
            self.addstr(
                ' %s' % (state, ),
                to_curses_attr(get_theme().COLOR_INFORMATION_BAR),
            )


class DynamicConversationInfoWin(ConversationInfoWin):
    __slots__ = ()

    def write_contact_jid(self, jid: JID) -> None:
        """
        Just displays the resource in an other color
        """
        log.debug("write_contact_jid DynamicConversationInfoWin, jid: %s",
                  jid.resource)
        theme = get_theme()
        color = to_curses_attr(theme.COLOR_INFORMATION_BAR)
        self.addstr('[', color)
        self.addstr(jid.bare,
                    to_curses_attr(theme.COLOR_CONVERSATION_NAME))
        if jid.resource:
            self.addstr(
                "/%s" % (jid.resource, ),
                to_curses_attr(theme.COLOR_CONVERSATION_RESOURCE))
        self.addstr('] ', color)


class MucInfoWin(InfoWin):
    """
    The line just above the information window, displaying information
    about the MUC we are viewing
    """

    __slots__ = ()

    def refresh(
        self,
        room: MucTab,
        window: Optional[TextWin] = None,
        user: Optional[User] = None,
        information: Optional[Dict[str, Any]] = None
    ) -> None:
        log.debug('Refresh: %s', self.__class__.__name__)
        self._win.erase()
        self.write_room_name(room)
        self.write_participants_number(room)
        self.write_own_nick(room)
        self.write_disconnected(room)
        self.write_role(user)
        if information:
            self.write_additional_information(information, room.name)
        if window:
            self.print_scroll_position(window)
        self.finish_line(get_theme().COLOR_INFORMATION_BAR)
        self._refresh()

    def write_additional_information(self, information: dict[str, AdditionalCb], jid: JID | str) -> None:
        """
        Write all information added by plugins by getting the
        value returned by the callbacks.
        """
        color = to_curses_attr(get_theme().COLOR_INFORMATION_BAR)
        for plugin in information.values():
            self.addstr(plugin(str(jid)), color)

    def write_room_name(self, room: 'MucTab') -> None:
        theme = get_theme()
        color = to_curses_attr(theme.COLOR_INFORMATION_BAR)
        label_color: TextStyle = theme.COLOR_GROUPCHAT_NAME

        if config.get('autocolor_tab_names'):
            label_color = ccg_text_to_color(
                theme.ccg_palette,
                room.jid.bare,
            ), -1, theme.MODE_TAB_NAME

        self.addstr('[', color)
        self.addstr(room.name,
                    to_curses_attr(label_color))
        self.addstr(']', color)

    def write_participants_number(self, room: 'MucTab') -> None:
        theme = get_theme()
        color = to_curses_attr(theme.COLOR_INFORMATION_BAR)
        self.addstr('{', color)
        self.addstr(
            str(len(room.users)),
            to_curses_attr(theme.COLOR_GROUPCHAT_NAME))
        self.addstr('} ', color)

    def write_disconnected(self, room: 'MucTab') -> None:
        """
        Shows a message if the room is not joined
        """
        if not room.joined:
            self.addstr(' -!- Not connected ',
                        to_curses_attr(get_theme().COLOR_INFORMATION_BAR))

    def write_own_nick(self, room: 'MucTab') -> None:
        """
        Write our own nick in the info bar
        """
        nick = room.own_nick
        if not nick:
            return
        self.addstr(
            truncate_nick(nick, 13),
            to_curses_attr(get_theme().COLOR_INFORMATION_BAR))

    def write_role(self, user: 'User' | None) -> None:
        """
        Write our own role and affiliation
        """
        if not user:
            return
        txt = ' ('
        if user.affiliation != 'none':
            txt += user.affiliation + ', '
        txt += user.role + ')'
        self.addstr(txt, to_curses_attr(get_theme().COLOR_INFORMATION_BAR))


class ConversationStatusMessageWin(InfoWin):
    """
    The upper bar displaying the status message of the contact
    """

    __slots__ = ()

    def refresh(self, jid: JID | str, contact: Contact | None) -> None:
        log.debug('Refresh: %s', self.__class__.__name__)
        try:
            jid = JID(jid)
        except InvalidJID:
            jid = JID('')
        if contact:
            if jid.resource:
                resource = contact[jid.full]
            else:
                resource = contact.get_highest_priority_resource()
        else:
            resource = None
        self._win.erase()
        if resource:
            self.write_status_message(resource)
        self.finish_line(get_theme().COLOR_INFORMATION_BAR)
        self._refresh()

    def write_status_message(self, resource: Resource) -> None:
        self.addstr(resource.status,
                    to_curses_attr(get_theme().COLOR_INFORMATION_BAR))


class BookmarksInfoWin(InfoWin):
    __slots__ = ()

    def refresh(self, preferred: str) -> None:
        log.debug('Refresh: %s', self.__class__.__name__)
        self._win.erase()
        self.write_remote_status(preferred)
        self.finish_line(get_theme().COLOR_INFORMATION_BAR)
        self._refresh()

    def write_remote_status(self, preferred: str) -> None:
        self.addstr('Remote storage: %s' % preferred,
                    to_curses_attr(get_theme().COLOR_INFORMATION_BAR))


class ConfirmStatusWin(Win):
    __slots__ = ('text', 'critical')

    def __init__(self, text: str, critical: bool = False) -> None:
        Win.__init__(self)
        self.text = text
        self.critical = critical

    def refresh(self) -> None:
        log.debug('Refresh: %s', self.__class__.__name__)
        self._win.erase()
        theme = get_theme()
        color: TextStyle
        if self.critical:
            color = theme.COLOR_WARNING_PROMPT
        else:
            color = theme.COLOR_INFORMATION_BAR
        c_color = to_curses_attr(color)
        self.addstr(self.text, c_color)
        self.finish_line(color)
        self._refresh()



# Copyright 2010-2011 Florent Le Coz <louiz@louiz.org>
#
# This file is part of Poezio.
#
# Poezio is free software: you can redistribute it and/or modify
# it under the terms of the GPL-3.0+ license. See the COPYING file.
"""
Timed events are the standard way to schedule events for later in poezio.

Once created, they must be added to the list of checked events with
:py:func:`Core.add_timed_event` (within poezio) or with
:py:func:`.PluginAPI.add_timed_event` (within a plugin).
"""

from datetime import datetime
from asyncio import Handle
from typing import Callable, ParamSpec, TypeVar


P = ParamSpec('P')
U = TypeVar('U')


class DelayedEventError(ValueError):
    def __init__(self) -> None:
        super().__init__("No kwargs are allowed for delayed events.")


class DelayedEvent:
    """
    A TimedEvent, but with the date calculated from now + a delay in seconds.
    Use it if you want an event to happen in, e.g. 6 seconds.
    """

    def __init__(self, delay: int | float, callback: Callable[P, U],
                 *args: P.args, **kwargs: P.kwargs) -> None:
        """
        Create a new DelayedEvent.

        :param int delay: The number of seconds.
        :param function callback: The handler that will be executed.
        :param args: Optional arguments passed to the handler.
        """
        self.callback = callback
        self.args = args
        self.delay = delay
        # An asyncio handler, as returned by call_later() or call_at()
        self.handler: Handle | None = None
        if kwargs:
            # asyncio’s call_later does not take kwargs
            raise DelayedEventError()


class TimedEvent(DelayedEvent):
    """
    An event with a callback that is called when the specified time is passed.

    The callback and its arguments should be passed as the lasts arguments.
    """

    def __init__(self, date: datetime, callback: Callable[P, U], *args: P.args,
                 **kwargs: P.kwargs) -> None:
        """
        Create a new timed event.

        :param datetime.datetime date: Time at which the callback must be run.
        :param function callback: The handler that will be executed.
        :param args: Optional arguments passed to the handler.
        """
        delta = date - datetime.now()
        delay = delta.total_seconds()
        DelayedEvent.__init__(self, delay, callback, *args, **kwargs)



"""
This plugin adds a "cyber" prefix to a random word in your chatroom messages.

Usage
-----

Say something in a MUC tab.

Configuration options
---------------------

.. glossary::

    frequency
        **Default:** ``10``

        The percentage of the time the plugin will activate (randomly). 100 for every message, <= 0 for never.
"""

from poezio.plugin import BasePlugin
from poezio import tabs
from random import choice, randint
from slixmpp import Message
import re

DEFAULT_CONFIG = {'cyber': {'frequency': 10}}


class Plugin(BasePlugin):

    default_config = DEFAULT_CONFIG

    def init(self) -> None:
        self.api.add_event_handler('muc_say', self.cyberize)

    def cyberize(self, msg: Message, tab: tabs.ChatTab) -> None:
        if randint(1, 100) > self.config.get('frequency'):
            return
        words = [
            word for word in re.split(r'\W+', msg['body']) if len(word) > 3
        ]
        if words:
            word = choice(words)
            msg['body'] = msg['body'].replace(word, 'cyber' + word)



"""
    Query and control an archive of messages stored on a server using
    XEP-0313: Message Archive Management(MAM).
"""

from __future__ import annotations

import logging
from datetime import datetime, timedelta, timezone
from hashlib import md5
from typing import (
    AsyncIterable,
    List,
    Optional,
    TYPE_CHECKING,
    TypedDict,
)

from slixmpp import JID, Message as SMessage
from slixmpp.exceptions import IqError, IqTimeout
from poezio.theming import get_theme, TextStyle
from poezio import tabs
from poezio import colors
from poezio.common import to_utc
from poezio.ui.types import (
    Message,
)

if TYPE_CHECKING:
    from poezio.core.core import Core


log = logging.getLogger(__name__)


class DiscoInfoException(Exception):
    pass


class MAMQueryException(Exception):
    def __init__(self) -> None:
        super().__init__("Query interrupted")


class NoMAMSupportException(Exception):
    pass


def make_line(
        tab: tabs.ChatTab,
        text: str,
        time: datetime,
        jid: JID,
        identifier: str = '',
        nick: str = ''
) -> Message:
    """Adds a textual entry in the TextBuffer"""

    # Convert to local timezone
    time = time.replace(tzinfo=timezone.utc).astimezone(tz=None)
    time = time.replace(tzinfo=None)
    color: TextStyle

    if isinstance(tab, tabs.MucTab):
        nick = jid.resource
        user = tab.get_user_by_name(nick)
        if user:
            color = user.color
        else:
            theme = get_theme()
            if theme.ccg_palette:
                fg_color = colors.ccg_text_to_color(theme.ccg_palette, nick)
                color = fg_color, -1
            else:
                mod = len(theme.LIST_COLOR_NICKNAMES)
                nick_pos = int(md5(nick.encode('utf-8')).hexdigest(), 16) % mod
                color = theme.LIST_COLOR_NICKNAMES[nick_pos]
    else:
        if jid.bare == tab.core.xmpp.boundjid.bare:
            if not nick:
                nick = tab.core.own_nick
            color = get_theme().COLOR_OWN_NICK
        else:
            color = get_theme().COLOR_REMOTE_USER
            if not nick:
                nick = tab.get_nick()
    return Message(
        txt=text,
        identifier=identifier,
        time=time,
        nickname=nick,
        nick_color=color,
        history=True,
        user=None,
    )


class IterateArgs(TypedDict, total=False):
    reverse: bool
    rsm: dict[str, int]
    total: int
    start: str | None
    end: str | None
    jid: JID
    with_jid: JID


async def get_mam_iterator(
        core: 'Core',
        groupchat: bool,
        remote_jid: JID,
        amount: int,
        reverse: bool = True,
        start: Optional[str] = None,
        end: Optional[str] = None,
        before: Optional[str] = None,
) -> AsyncIterable[SMessage]:
    """Get an async iterator for this mam query"""
    try:
        query_jid = remote_jid if groupchat else JID(core.xmpp.boundjid.bare)
        iq = await core.xmpp.plugin['xep_0030'].get_info(jid=query_jid)
    except (IqError, IqTimeout) as exc:
        raise DiscoInfoException() from exc
    if 'urn:xmpp:mam:2' not in iq['disco_info'].get_features():
        raise NoMAMSupportException()

    args: IterateArgs = {
        'reverse': reverse,
        'rsm': {
            'max': 30,
        },
    }

    if groupchat:
        args['jid'] = remote_jid
    else:
        args['with_jid'] = remote_jid

    args['total'] = amount
    args['start'] = start
    args['end'] = end
    return core.xmpp['xep_0313'].iterate(**args)  # type: ignore


class MessageDict(TypedDict):
    time: datetime
    jid: JID
    text: str
    identifier: str


def _parse_message(msg: SMessage) -> MessageDict:
    """Parse info inside a MAM forwarded message"""
    forwarded = msg['mam_result']['forwarded']
    message = forwarded['stanza']
    result: MessageDict = {
        'time': forwarded['delay']['stamp'],
        'jid': message['from'],
        'text': message['body'],
        'identifier': message['id']
    }
    return result


def _ignore_private_message(stanza: SMessage, filter_jid: Optional[JID]) -> bool:
    """Returns True if a MUC-PM should be ignored, as prosody returns
    all PMs within the same room.
    """
    if filter_jid is None:
        return False
    sent = stanza['from'].bare != filter_jid.bare
    if sent and stanza['to'].full != filter_jid.full:
        return True
    elif not sent and stanza['from'].full != filter_jid.full:
        return True
    return False


async def retrieve_messages(tab: tabs.ChatTab,
                            results: AsyncIterable[SMessage],
                            amount: int = 100) -> List[Message]:
    """Run the MAM query and put messages in order"""
    msg_count = 0
    to_add: list[Message] = []
    tab_is_private = isinstance(tab, tabs.PrivateTab)
    filter_jid = None
    if tab_is_private:
        filter_jid = tab.jid
    try:
        async for msg in results:
            stanza = msg['mam_result']['forwarded']['stanza']
            if stanza.xml.find('{%s}%s' % ('jabber:client', 'body')) is not None:
                if _ignore_private_message(stanza, filter_jid):
                    continue
                args = _parse_message(msg)
                to_add.append(make_line(tab, **args))
                msg_count += 1
                if msg_count == amount:
                    to_add.reverse()
                    return to_add
    except (IqError, IqTimeout) as exc:
        log.debug('Unable to complete MAM query: %s', exc, exc_info=True)
        raise MAMQueryException() from exc
    to_add.reverse()
    return to_add


async def fetch_history(tab: tabs.ChatTab,
                        start: Optional[datetime] = None,
                        end: Optional[datetime] = None,
                        amount: int = 100) -> List[Message]:
    remote_jid = tab.jid
    if not end:
        for msg in tab._text_buffer.messages:
            if isinstance(msg, Message):
                end = msg.time
                end -= timedelta(microseconds=1)
                break
    if end is None:
        end = datetime.now() + timedelta(seconds=1)
    end = to_utc(end)
    end_str = datetime.strftime(end, '%Y-%m-%dT%H:%M:%SZ')

    start_str = None
    if start is not None:
        start = to_utc(start)
        start_str = datetime.strftime(start, '%Y-%m-%dT%H:%M:%SZ')

    mam_iterator = await get_mam_iterator(
        core=tab.core,
        groupchat=isinstance(tab, tabs.MucTab),
        remote_jid=remote_jid,
        amount=amount,
        end=end_str,
        start=start_str,
        reverse=True,
    )
    return await retrieve_messages(tab, mam_iterator, amount)



#!/usr/bin/env python3
# Copyright 2011 Florent Le Coz <louiz@louiz.org>
#
# This file is part of Poezio.
#
# Poezio is free software: you can redistribute it and/or modify
# it under the terms of the GPL-3.0+ license. See the COPYING file.
"""
This file is a standalone program that reads commands on
stdin and executes them (each line should be a command).

Usage: cat some_fifo | ./daemon.py

Poezio writes commands in the fifo, and this daemon executes them on the
local machine.
Note that you should not start this daemon if you do not trust the remote
machine that is running poezio, since this could make it run any (dangerous)
command on your local machine.
"""

import sys
import threading
import subprocess
import shlex
import logging

from subprocess import DEVNULL
from typing import IO

log = logging.getLogger(__name__)


class Executor(threading.Thread):
    """
    Just a class to execute commands in a thread.  This way, the execution
    can totally fail, we don’t care, and we can start commands without
    having to wait for them to return.
    WARNING: Be careful to properly escape what is untrusted by using
    shlex.quote for example.
    """

    def __init__(self, command: list[str], remote: bool = False) -> None:
        threading.Thread.__init__(self)
        self.command = command
        self.remote = remote
        # check for > or >> special case
        self.filename = None
        self.redirection_mode = 'w'
        if len(command) >= 3:
            if command[-2] in ('>', '>>'):
                self.filename = command.pop(-1)
                if command[-1] == '>>':
                    self.redirection_mode = 'a'
                command.pop(-1)

    def run(self) -> None:
        log.debug('executing %s', self.command)
        stdout: IO[str] | int = DEVNULL
        if self.filename:
            try:
                stdout = open(self.filename, self.redirection_mode)
            except OSError:
                log.exception(
                    'Could not open redirection file: %s',
                    self.filename,
                )
                return
        try:
            subprocess.call(self.command, stdout=stdout, stderr=DEVNULL)
        except Exception:
            if self.remote:
                import traceback
                print(traceback.format_exc())
            else:
                log.exception('Could not execute %s:', self.command)


def main() -> None:
    while True:
        line = sys.stdin.readline()
        if line == '':
            break
        command = shlex.split(line)
        e = Executor(command, remote=True)
        e.start()


if __name__ == '__main__':
    main()



"""
Defines the global config instance, used to get or set (and save) values
from/to the config file.

This module has the particularity that some imports and global variables
are delayed because it would mean doing an incomplete setup of the python
loggers.

TODO: get http://bugs.python.org/issue1410680 fixed, one day, in order
to remove our ugly custom I/O methods.
"""

import logging
import logging.config
import os
import sys

from configparser import RawConfigParser, NoOptionError, NoSectionError
from pathlib import Path
from typing import Dict, List, Optional, Union, cast, Any, TypedDict

from poezio import xdg
from slixmpp import JID, InvalidJID

log = logging.getLogger(__name__)  # type: logging.Logger

ConfigValue = Union[str, int, float, bool]

ConfigDict = Dict[str, Dict[str, ConfigValue]]

USE_DEFAULT_SECTION = '__DEFAULT SECTION PLACEHOLDER__'

DEFAULT_CONFIG: ConfigDict = {
    'Poezio': {
        'ack_message_receipts': True,
        'add_space_after_completion': True,
        'after_completion': ',',
        'alternative_nickname': '',
        'auto_reconnect': True,
        'autocolor_tab_names': False,
        'autorejoin_delay': '5',
        'autorejoin': False,
        'beep_on': 'highlight private invite disconnect',
        'ca_cert_path': '',
        'certificate': '',
        'certfile': '',
        'ciphers': 'HIGH+kEDH:HIGH+kEECDH:HIGH:!PSK:!SRP:!3DES:!aNULL',
        'connection_check_interval': 300,
        'connection_timeout_delay': 30,
        'create_gaps': False,
        'custom_host': '',
        'custom_port': '',
        'default_nick': '',
        'default_muc_service': '',
        'device_id': '',
        'nick_color_aliases': True,
        'display_activity_notifications': False,
        'display_gaming_notifications': False,
        'display_mood_notifications': False,
        'display_tune_notifications': False,
        'display_user_color_in_join_part': True,
        'enable_avatars': True,
        'enable_carbons': True,
        'enable_reactions': True,
        'enable_css_parsing': True,
        'enable_user_activity': True,
        'enable_user_gaming': True,
        'enable_user_mood': True,
        'enable_user_nick': True,
        'enable_user_tune': True,
        'enable_vertical_tab_list': True,
        'enable_xhtml_im': True,
        'enable_smacks': False,
        'eval_password': '',
        'exec_remote': False,
        'extract_inline_images': True,
        'filter_info_messages': '',
        'force_encryption': True,
        'go_to_previous_tab_on_alt_number': False,
        'group_corrections': True,
        'hide_exit_join': -1,
        'hide_status_change': 120,
        'hide_user_list': False,
        'highlight_on': '',
        'ignore_certificate': False,
        'ignore_private': False,
        'image_use_half_blocks': False,
        'information_buffer_popup_on': 'error roster warning help info',
        'information_buffer_type_filter': '',
        'jid': '',
        'keyfile': '',
        'lang': 'en',
        'lazy_resize': True,
        'log_dir': '',
        'log_errors': True,
        'mam_sync': True,
        'mam_sync_limit': 2000,
        'max_lines_in_memory': 2048,
        'max_messages_in_memory': 2048,
        'max_nick_length': 25,
        'muc_history_length': 50,
        'notify_messages': True,
        'open_all_bookmarks': False,
        'password': '',
        'plugins_autoload': '',
        'plugins_conf_dir': '',
        'plugins_dir': '',
        'popup_time': 4,
        'private_auto_response': '',
        'remote_fifo_path': './',
        'request_message_receipts': True,
        'rooms': '',
        'roster_group_sort': 'name',
        'roster_show_offline': False,
        'roster_sort': 'jid:show',
        'save_status': True,
        'self_ping_interval': 600,
        'self_ping_timeout': 60,
        'send_chat_states': True,
        'send_initial_presence': True,
        'send_os_info': True,
        'send_poezio_info': True,
        'send_time': True,
        'separate_history': False,
        'server': 'anon.jeproteste.info',
        'show_composing_tabs': 'direct',
        'show_inactive_tabs': True,
        'show_jid_in_conversations': True,
        'show_muc_jid': False,
        'show_roster_jids': True,
        'show_roster_subscriptions': '',
        'show_s2s_errors': True,
        'show_tab_names': False,
        'show_tab_numbers': True,
        'show_timestamps': True,
        'show_useless_separator': True,
        'status': '',
        'status_message': '',
        'synchronise_open_rooms': True,
        'theme': 'default',
        'themes_dir': '',
        'tmp_image_dir': '',
        'unique_prefix_tab_names': False,
        'use_bookmarks_method': '',
        'use_log': True,
        'use_remote_bookmarks': True,
        'user_list_sort': 'desc',
        'use_tab_nicks': True,
        'tls_verify_cert': True,
        'vertical_tab_list_size': 20,
        'vertical_tab_list_sort': 'desc',
        'whitespace_interval': 300,
        'words': ''
    },
    'bindings': {
        'M-i': '^I'
    },
    'var': {
        'folded_roster_groups': '',
        'info_win_height': 2
    },
    'muc_colors': {}
}


class PoezioConfigParser(RawConfigParser):
    def optionxform(self, value: Any) -> str:
        return str(value)


class Config:
    """
    load/save the config to a file
    """

    configparser: PoezioConfigParser
    file_name: Path
    default: ConfigDict
    default_section: str = 'Poezio'

    def __init__(self, file_name: Path, default: ConfigDict | None = None) -> None:
        self.configparser = PoezioConfigParser()
        # This resolve() call is needed to make sure we follow symlinks if they
        # exist, instead of replacing the symlink with a new file on write.
        self.file_name = file_name.resolve()
        self.read_file()
        self.default = default or {}

    def optionxform(self, value: int | float | str | None) -> str:
        return str(value)

    def read_file(self) -> None:
        self.configparser.read(self.file_name, encoding='utf-8')
        # Check config integrity and fix it if it’s wrong
        # only when the object is the main config
        if self.__class__ is Config:
            for section in ('bindings', 'var'):
                if not self.has_section(section):
                    self.add_section(section)

    def get(self,
            option: str,
            default: Optional[ConfigValue] = None,
            section: str = USE_DEFAULT_SECTION) -> Any:
        """
        get a value from the config but return
        a default value if it is not found
        The type of default defines the type
        returned
        """
        if section == USE_DEFAULT_SECTION:
            section = self.default_section
        if default is None:
            default = self.default.get(section, {}).get(option, '')

        res: Optional[ConfigValue]
        try:
            if isinstance(default, bool):
                res = self.configparser.getboolean(section, option)
            elif isinstance(default, int):
                res = self.configparser.getint(section, option)
            elif isinstance(default, float):
                res = self.configparser.getfloat(section, option)
            else:
                res = self.configparser.get(section, option)
        except (NoOptionError, NoSectionError, ValueError, AttributeError):
            return default

        if res is None:
            return default
        return res

    def _get_default(self, option: str, section: str) -> Any:
        if self.default:
            return self.default.get(section, {}).get(option)
        else:
            return ''

    def sections(self) -> list[str]:
        return self.configparser.sections()

    def options(self, section: str = '') -> list[str]:
        return self.configparser.options(section)

    def has_option(self, section: str, option: str) -> bool:
        return self.configparser.has_option(section, option)

    def has_section(self, section: str) -> bool:
        return self.configparser.has_section(section)

    def add_section(self, section: str) -> None:
        return self.configparser.add_section(section)

    def remove_section(self, section: str) -> bool:
        return self.configparser.remove_section(section)

    def get_by_tabname(self,
                       option: str,
                       tabname: JID,
                       fallback: bool = True,
                       fallback_server: bool = True,
                       default: str | int | float = ''):
        """
        Try to get the value for the option. First we look in
        a section named `tabname`, if the option is not present
        in the section, we search for the global option if fallback is
        True. And we return `default` as a fallback as a last resort.
        """
        if self.default and (not default) and fallback:
            default = self.default.get(self.default_section, {}).get(option, '')
        if tabname in self.sections():
            if option in self.options(str(tabname)):
                # We go the tab-specific option
                return self.get(option, default, tabname.full)
        if fallback_server:
            return self.get_by_servname(tabname, option, default, fallback)
        if fallback:
            # We fallback to the global option
            return self.get(option, default)
        return default

    def get_by_servname(self, jid: str | JID, option: str, default: str | int | float,
                        fallback: bool = True):
        """
        Try to get the value of an option for a server
        """
        try:
            server = JID(jid).server
        except InvalidJID:
            server = ''
        if server:
            server = '@' + server
            if server in self.sections() and option in self.options(server):
                return self.get(option, default, server)
        if fallback:
            return self.get(option, default)
        return default

    def getstr(self, option: str, section: str = USE_DEFAULT_SECTION) -> str:
        """
        get a value and returns it as a string
        """
        if section == USE_DEFAULT_SECTION:
            section = self.default_section
        try:
            return self.configparser.get(section, option)
        except (NoOptionError, NoSectionError, ValueError, AttributeError):
            return cast(str, self._get_default(option, section))

    def getint(self, option: str, section: str = USE_DEFAULT_SECTION) -> int:
        """
        get a value and returns it as an int
        """
        if section == USE_DEFAULT_SECTION:
            section = self.default_section
        try:
            return self.configparser.getint(section, option)
        except (NoOptionError, NoSectionError, ValueError, AttributeError):
            return cast(int, self._get_default(option, section))

    def getfloat(self, option: str, section: str = USE_DEFAULT_SECTION) -> float:
        """
        get a value and returns it as a float
        """
        if section == USE_DEFAULT_SECTION:
            section = self.default_section
        try:
            return self.configparser.getfloat(section, option)
        except (NoOptionError, NoSectionError, ValueError, AttributeError):
            return cast(float, self._get_default(option, section))

    def getbool(self, option: str, section: str = USE_DEFAULT_SECTION) -> bool:
        """
        get a value and returns it as a boolean
        """
        if section == USE_DEFAULT_SECTION:
            section = self.default_section
        try:
            return self.configparser.getboolean(section, option)
        except (NoOptionError, NoSectionError, ValueError, AttributeError):
            return cast(bool, self._get_default(option, section))

    def getlist(self, option: str, section: str = USE_DEFAULT_SECTION) -> list[str]:
        if section == USE_DEFAULT_SECTION:
            section = self.default_section
        return self.getstr(option, section).split(':')

    def set_and_save(self, option: str, value: ConfigValue,
                     section: str = USE_DEFAULT_SECTION) -> tuple[str, str]:
        """
        set the value in the configuration then save it
        to the file
        """
        # Special case for a 'toggle' value. We take the current value
        # and set the opposite. Warning if the no current value exists
        # or it is not a bool.
        if section == USE_DEFAULT_SECTION:
            section = self.default_section
        if isinstance(value, str) and value == "toggle":
            current = self.getbool(option, section)
            if isinstance(current, bool):
                value = str(not current).lower()
            else:
                if current.lower() == "false":
                    value = "true"
                elif current.lower() == "true":
                    value = "false"
                else:
                    return (
                        'Could not toggle option: %s.'
                        ' Current value is %s.' % (option, current or "empty"),
                        'Warning')
        value = str(value)
        if self.has_section(section):
            self.configparser.set(section, option, value)
        else:
            self.add_section(section)
            self.configparser.set(section, option, value)
        if not write_in_file(self.file_name, section, option, value):
            return ('Unable to write in the config file', 'Error')
        if isinstance(option, str) and 'password' in option and 'eval_password' not in option:
            value = '********'
        return ("%s=%s" % (option, value), 'Info')

    def remove_and_save(self, option: str,
                        section: str = USE_DEFAULT_SECTION) -> tuple[str, str]:
        """
        Remove an option and then save it the config file
        """
        if section == USE_DEFAULT_SECTION:
            section = self.default_section
        if self.has_section(section):
            self.configparser.remove_option(section, option)
        if not remove_in_file(self.file_name, section, option):
            return ('Unable to save the config file', 'Error')
        return ('Option %s deleted' % option, 'Info')

    def silent_set(self, option: str, value: ConfigValue,
                   section: str = USE_DEFAULT_SECTION) -> bool:
        """
        Set a value, save, and return True on success and False on failure
        """
        if section == USE_DEFAULT_SECTION:
            section = self.default_section
        if self.has_section(section):
            self.configparser.set(section, option, str(value))
        else:
            self.add_section(section)
            self.configparser.set(section, option, str(value))
        return write_in_file(self.file_name, section, option, str(value))

    def set(self, option: str, value: ConfigValue,
            section: str = USE_DEFAULT_SECTION) -> None:
        """
        Set the value of an option temporarily
        """
        if section == USE_DEFAULT_SECTION:
            section = self.default_section
        try:
            self.configparser.set(section, option, str(value))
        except NoSectionError:
            pass

    def to_dict(self) -> dict[str, dict[str, ConfigValue | None]]:
        """
        Returns a dict of the form {section: {option: value, option: value}, …}
        """
        res: dict[str, dict[str, ConfigValue | None]] = {}
        for section in self.sections():
            res[section] = {}
            for option in self.options(section):
                res[section][option] = self.get(option, "", section)
        return res


def write_in_file(file_name: Path, section: str, option: str,
                  value: ConfigValue) -> bool:
    """
    Our own way to save write the value in the file
    Just find the right section, and then find the
    right option, and edit it.
    """
    result = parse_file(file_name)
    if not result:
        return False
    else:
        sections, result_lines = result

    if section not in sections:
        result_lines.append('[%s]' % section)
        result_lines.append('%s = %s' % (option, value))
    else:
        begin, end = sections[section]
        pos = find_line(result_lines, begin, end, option)

        if pos == -1:
            result_lines.insert(end, '%s = %s' % (option, value))
        else:
            result_lines[pos] = '%s = %s' % (option, value)

    return write_entire_file(file_name, result_lines)


def remove_in_file(file_name: Path, section: str, option: str) -> bool:
    """
    Our own way to remove an option from the file.
    """
    result = parse_file(file_name)
    if not result:
        return False
    else:
        sections, result_lines = result

    if section not in sections:
        log.error(
            'Tried to remove the option %s from a non-'
            'existing section (%s)', option, section)
        return True
    else:
        begin, end = sections[section]
        pos = find_line(result_lines, begin, end, option)

        if pos == -1:
            log.error(
                'Tried to remove a non-existing option %s'
                ' from section %s', option, section)
            return True
        else:
            del result_lines[pos]

    return write_entire_file(file_name, result_lines)


def write_entire_file(file_name: Path, lines: list[str]) -> bool:
    """
    Write the config file, write to a temporary file
    before copying it to the final destination
    """
    try:
        filename = file_name.parent / f'.{file_name.name}.tmp'
        with os.fdopen(
                os.open(filename, os.O_WRONLY | os.O_CREAT, 0o600),
                'w',
                encoding='utf-8') as fd:
            for line in lines:
                fd.write('%s\n' % line)
        filename.replace(file_name)
    except Exception:
        success = False
        log.exception('Unable to save the config file.')
    else:
        success = True
    return success


def parse_file(file_name: Path) -> tuple[dict[str, list[int]], list[str]] | None:
    """
    Parse the config file and return the list of sections with
    their start and end positions, and the lines in the file.

    Duplicate sections are preserved but ignored for the parsing.

    Returns an empty tuple if reading fails
    """
    if file_ok(file_name):
        try:
            with file_name.open('r', encoding='utf-8') as df:
                lines_before: List[str] = [line.strip() for line in df]
        except OSError:
            log.exception(
                'Unable to read the config file %s',
                file_name,
            )
            return None
    else:
        lines_before = []

    sections: Dict[str, List[int]] = {}
    duplicate_section = False
    current_section = ''
    current_line = 0

    for line in lines_before:
        if line.startswith('['):
            if not duplicate_section and current_section:
                sections[current_section][1] = current_line

            duplicate_section = False
            current_section = line[1:-1]

            if current_section in sections:
                log.error('Error while reading the configuration file,'
                          ' skipping until next section')
                duplicate_section = True
            else:
                sections[current_section] = [current_line, current_line]

        current_line += 1
    if not duplicate_section and current_section:
        sections[current_section][1] = current_line

    return (sections, lines_before)


def find_line(lines: List[str], start: int, end: int, option: str) -> int:
    """
    Get the number of the line containing the option in the
    relevant part of the config file.

    Returns -1 if the option isn’t found
    """
    current = start
    for line in lines[start:end]:
        if (line.startswith('%s ' % option)
                or line.startswith('%s=' % option)):
            return current
        current += 1
    return -1


def file_ok(filepath: Path) -> bool:
    """
    Returns True if the file exists and is readable and writable,
    False otherwise.
    """
    val = filepath.exists()
    val &= os.access(filepath, os.R_OK | os.W_OK)
    return bool(val)


def get_image_cache() -> Optional[Path]:
    if not config.get('extract_inline_images'):
        return None
    tmp_dir = config.getstr('tmp_image_dir')
    if tmp_dir:
        return Path(tmp_dir)
    return xdg.CACHE_HOME / 'images'


def check_config() -> None:
    """
    Check the config file and print results
    """
    class ResultDict(TypedDict):
        missing: list[str]
        changed: list[tuple[str, Any, Any]]

    result: ResultDict = {'missing': [], 'changed': []}
    for option in DEFAULT_CONFIG['Poezio']:
        value = config.get(option)
        if value != DEFAULT_CONFIG['Poezio'][option]:
            result['changed'].append((option, value,
                                      DEFAULT_CONFIG['Poezio'][option]))
        else:
            value = config.get(option, default='')
            upper = value.upper()
            default = str(DEFAULT_CONFIG['Poezio'][option]).upper()
            if upper != default:
                result['missing'].append(option)

    result['changed'].sort(key=lambda x: x[0])
    result['missing'].sort()
    if result['changed']:
        print(
            '\033[1mOptions changed from the default configuration:\033[0m\n')
        for option, new_value, default in result['changed']:
            print(
                '    \033[1m%s\033[0m = \033[33m%s\033[0m (default: \033[32m%s\033[0m)'
                % (option, new_value, default))

    if result['missing']:
        print('\n\033[1mMissing options:\033[0m (the defaults are used)\n')
        for option in result['missing']:
            print('    \033[31m%s\033[0m' % option)


def create_global_config(filename: Path) -> None:
    "Create the global config object, or crash"
    try:
        global config
        config = Config(filename, DEFAULT_CONFIG)
    except Exception:
        import traceback
        sys.stderr.write('Poezio was unable to read or'
                         ' parse the config file.\n')
        traceback.print_exc(limit=0)
        sys.exit(1)


def setup_logging(debug_file: str = '') -> None:
    "Change the logging config according to the cmdline options and config"
    global LOG_DIR
    LOG_DIR = config.get('log_dir')
    LOG_DIR = Path(LOG_DIR).expanduser() if LOG_DIR else xdg.DATA_HOME / 'logs'
    from copy import deepcopy
    logging_config = deepcopy(LOGGING_CONFIG)
    if config.get('log_errors'):
        try:
            LOG_DIR.mkdir(parents=True, exist_ok=True)
        except OSError:
            # We can’t really log any error here, because logging isn’t setup yet.
            pass
        else:
            logging_config['root']['handlers'].append('error')
            logging_config['handlers']['error'] = {
                'level': 'ERROR',
                'class': 'logging.FileHandler',
                'filename': str(LOG_DIR / 'errors.log'),
                'formatter': 'simple',
            }
            logging.disable(logging.WARNING)

    if debug_file:
        logging_config['root']['handlers'].append('debug')
        logging_config['handlers']['debug'] = {
            'level': 'DEBUG',
            'class': 'logging.FileHandler',
            'filename': debug_file,
            'formatter': 'simple',
        }
        logging.disable(logging.NOTSET)

    if logging_config['root']['handlers']:
        logging.config.dictConfig(cast(dict[str, Any], logging_config))
    else:
        logging.disable(logging.ERROR)
        logging.basicConfig(level=logging.CRITICAL)


class RootConfig(TypedDict):
    handlers: list[str]
    propagate: bool
    level: str


class LoggingConfig(TypedDict):
    version: int
    disable_existing_loggers: bool
    formatters: dict[str, dict[str, str]]
    handlers: dict[str, dict[str, str]]
    root: RootConfig


LOGGING_CONFIG: LoggingConfig = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'simple': {
            'format': '%(asctime)s %(levelname)s:%(module)s:%(message)s'
        }
    },
    'handlers': {},
    'root': {
        'handlers': [],
        'propagate': True,
        'level': 'DEBUG',
    }
}

# Global config object. Is setup for real in poezio.py
config = Config(Path('/dev/null'))

# the global log dir
LOG_DIR = Path()



from poezio.plugin import BasePlugin
from poezio import tabs
from slixmpp import Message


class Plugin(BasePlugin):
    def init(self) -> None:
        self.api.add_command('plugintest', self.command_plugintest,
                             'Test command')
        self.api.add_tab_command(tabs.MucTab, 'plugintest',
                                 self.command_tab_plugintest, 'Test command')
        self.api.add_slix_event_handler('message', self.on_message)
        self.api.information("Plugin loaded")

    def cleanup(self) -> None:
        self.api.information("Plugin unloaded")

    def on_message(self, message: Message) -> None:
        self.api.information(
            "Test plugin received message: {}".format(message))

    def command_tab_plugintest(self, args: str) -> None:
        self.api.information("Command for MucTabs! With args {}".format(args))
        self.api.del_tab_command(tabs.MucTab, 'plugintest')

    def command_plugintest(self, args: str) -> None:
        self.api.information("Command! With args {}".format(args))



"""
A MucListTab is a tab listing the rooms on a conference server.

It has no functionality except scrolling the list, and allowing the
user to join the rooms.
"""
import logging
from typing import Any, Callable, TYPE_CHECKING
from collections.abc import Iterator

from poezio.tabs import ListTab
from poezio.core.structs import Command

from slixmpp import JID, Iq
from slixmpp.plugins.xep_0030.stanza.items import DiscoItem

if TYPE_CHECKING:
    from poezio.core.core import Core

log = logging.getLogger(__name__)


class MucListTab(ListTab):
    """
    A tab listing rooms from a specific server, displaying various information,
    scrollable, and letting the user join them, etc
    """
    plugin_commands: dict[str, Command] = {}
    plugin_keys: dict[str, Callable[..., Any]] = {}

    def __init__(self, core: 'Core', server: JID) -> None:
        ListTab.__init__(self, core, server.full, "“j”: join room.",
                         'Chatroom list on server %s (Loading)' % server,
                         [('node-part', 0), ('name', 2), ('users', 3)])
        self.key_func['j'] = self.join_selected
        self.key_func['J'] = self.join_selected_no_focus
        self.key_func['^M'] = self.join_selected

    def get_columns_sizes(self) -> dict[str, int]:
        return {
            'node-part': int(self.width * 2 / 8),
            'name': int(self.width * 5 / 8),
            'users':
            self.width - int(self.width * 2 / 8) - int(self.width * 5 / 8)
        }

    def join_selected_no_focus(self) -> None:
        return

    def on_muc_list_item_received(self, iq: Iq) -> None:
        """
        Callback called when a disco#items result is received
        Used with command_list
        """
        if iq['type'] == 'error':
            self.set_error(iq['error']['type'], iq['error']['code'],
                           iq['error']['text'])
            return

        def get_items() -> Iterator[tuple[str, str, str]]:
            substanza = iq['disco_items']
            for item in substanza['substanzas']:
                if isinstance(item, DiscoItem):
                    yield (item['jid'], item['node'], item['name'])

        items = [(item[0].split('@')[0], item[0], item[2] or '', '')
                 for item in get_items()]
        items = sorted(items, key=lambda item: item[0])
        self.listview.set_lines(items)
        self.info_header.message = 'Chatroom list on server %s' % self.name
        if self.core.tabs.current_tab is self:
            self.refresh()
        else:
            self.state = 'highlight'
            self.refresh_tab_win()
        self.core.doupdate()

    def join_selected(self) -> None:
        row = self.listview.get_selected_row()
        if not row:
            return
        self.core.register_task(self.core.command.join(row[1]))



# Copyright 2018 Emmanuel Gil Peyrot <linkmauve@linkmauve.fr>
#
# This file is part of Poezio.
#
# Poezio is free software: you can redistribute it and/or modify
# it under the terms of the GPL-3.0+ license. See the COPYING file.
"""
Implements the XDG base directory specification.

https://specifications.freedesktop.org/basedir-spec/basedir-spec-latest.html
"""

from pathlib import Path
from os import environ
from typing import Dict

# $HOME has already been checked to not be None in test_env().
DEFAULT_PATHS: Dict[str, Path] = {
    'XDG_CONFIG_HOME': Path.home() / '.config',
    'XDG_DATA_HOME': Path.home() / '.local' / 'share',
    'XDG_CACHE_HOME': Path.home() / '.cache',
}


class BadXDGVariable(ValueError):
    def __init__(self) -> None:
        super().__init__("Invalid XDG basedir variable")


def _get_directory(variable: str) -> Path:
    """
    returns the default configuration directory path
    """
    if variable not in DEFAULT_PATHS:
        raise BadXDGVariable()
    xdg = environ.get(variable)
    if xdg is not None:
        xdg_path = Path(xdg)
        if xdg_path.is_absolute():
            return xdg_path / 'poezio'
    return DEFAULT_PATHS[variable] / 'poezio'


CONFIG_HOME = _get_directory('XDG_CONFIG_HOME')
DATA_HOME = _get_directory('XDG_DATA_HOME')
CACHE_HOME = _get_directory('XDG_CACHE_HOME')



"""
This plugin lets you perform simple replacements on the last message.

Usage
-----

.. note:: the ``/``, ``#``, ``!``, ``:`` and ``;`` chars can be used as separators,
         even if the examples only use ``/``


Regex replacement
~~~~~~~~~~~~~~~~~

Once the plugin is loaded, any message matching the following regex

.. code-block:: text

    ^s/(.+?)/(.*?)(/|/g)?$

will be interpreted as a regex replacement, and the substitution will be
applied to the last sent message.

For example, if you sent the message::

    This tab lists all public rooms on a MUC service. It is currently very limited but will be improved in the future. There currently is no way to search a room.

And you now want to replace “MUC” with “multi-user chat”, you input::

    s/MUC/multi-user chat

And poezio will correct the message for you.


Raw string replacement
~~~~~~~~~~~~~~~~~~~~~~

Once the plugin is loaded, any message matching the following regex

.. code-block:: text

    ^r/(.+?)/(.*?)(/|/g)?$

will be interpreted as a replacement, and the substitution will be applied
to the last send message.

This variant is useful if you don’t want to care about regular expressions
(and you do not want to have to escape stuff like space or backslashes).


"""

from poezio.plugin import BasePlugin
from poezio import tabs
from slixmpp import Message
import re

allowed_separators = '/#!:;'
sed_re = re.compile(
    '^([sr])(?P<sep>[%s])(.+?)(?P=sep)(.*?)((?P=sep)|(?P=sep)g)?$' %
    allowed_separators)


class Plugin(BasePlugin):
    def init(self) -> None:
        self.api.add_event_handler('muc_say', self.sed_fix)
        self.api.add_event_handler('conversation_say', self.sed_fix)
        self.api.add_event_handler('private_say', self.sed_fix)

    def sed_fix(self, msg: Message, tab: tabs.ChatTab) -> None:
        if not tab.last_sent_message:
            return
        if 'correct' not in tab.commands:
            return
        body = new_body = tab.last_sent_message['body']
        match = sed_re.match(msg['body'])
        if not match:
            return
        typ, sep, remove, put, matchall = match.groups()

        replace_all = False
        if matchall == sep + 'g':
            replace_all = True

        if typ == 's':
            try:
                if replace_all:
                    new_body = re.sub(remove, put, body)
                else:
                    new_body = re.sub(remove, put, body, count=1)
            except Exception as e:
                self.api.information(
                    'Invalid regex for the autocorrect '
                    'plugin: %s' % e, 'Error')
                return
        elif typ == 'r':
            if replace_all:
                new_body = body.replace(remove, put)
            else:
                new_body = body.replace(remove, put, 1)

        if body != new_body:
            msg['body'] = new_body
            msg['replace']['id'] = tab.last_sent_message['id']



"""
This plugin broadcasts a message to all your joined rooms.

.. note:: With great power comes great responsibility.
          Use with moderation.

Command
-------

.. glossary::

    /amsg
        **Usage:** ``/amsg <message>``

        Broadcast a message.


"""
from poezio.plugin import BasePlugin
from poezio.tabs import MucTab


class Plugin(BasePlugin):
    def init(self) -> None:
        self.api.add_command(
            'amsg',
            self.command_amsg,
            usage='<message>',
            short='Broadcast a message',
            help='Broadcast the message to all the joined rooms.')

    async def command_amsg(self, args: str) -> None:
        for room in self.core.tabs:
            if isinstance(room, MucTab) and room.joined:
                await room.command_say(args)



"""
Do a disco#info query on a JID

Usage
-----

.. glossary::

    /disco
        **Usage:** ``/disco <JID>``

        This command queries a JID for its disco#info.

        There is no cache, as this is generally used for debug more than
        anything user-related.
"""

from poezio.plugin import BasePlugin
from poezio.decorators import command_args_parser
from slixmpp import InvalidJID, Iq, JID
from slixmpp.exceptions import IqError, IqTimeout


class Plugin(BasePlugin):
    def init(self) -> None:
        self.api.add_command(
            'disco',
            self.command_disco,
            usage='<JID> [node] [info|items]',
            short='Get the disco#info of a JID',
            help='Get the disco#info of a JID')

    def on_info(self, iq: Iq) -> None:
        if iq['type'] == 'error':
            self.api.information(iq['error']['text'] or iq['error']['condition'], 'Error')
            return

        info = iq['disco_info']
        identities = (str(identity) for identity in info['identities'])
        self.api.information('\n'.join(identities), 'Identities')
        features = sorted(str(feature) for feature in info['features'])
        self.api.information('\n'.join(features), 'Features')
        title = 'Server Info'
        server_info = []
        for field in info['form']:
            var = field['var']
            if field['type'] == 'hidden' and var == 'FORM_TYPE':
                title = field['value'][0]
                continue
            sep = '\n  ' + len(var) * ' '
            field_value = field.get_value(convert=False)
            value = sep.join(field_value) if isinstance(field_value,
                                                        list) else field_value
            server_info.append('%s: %s' % (var, value))
        if server_info:
            self.api.information('\n'.join(server_info), title)

    def on_items(self, iq: Iq) -> None:
        if iq['type'] == 'error':
            self.api.information(iq['error']['text'] or iq['error']['condition'], 'Error')
            return

        def describe(item: tuple[str, str | None, str | None]) -> str:
            text = item[0]
            node = item[1]
            name = item[2]
            if node is not None:
                text += ', node=' + node
            if name is not None:
                text += ', name=' + name
            return text

        items = iq['disco_items']
        self.api.information('\n'.join(describe(item) for item in items['items']), 'Items')

    @command_args_parser.quoted(1, 3)
    async def command_disco(self, args: list[str]) -> None:
        if not args:
            self.core.command.help('disco')
            return
        if len(args) == 1:
            jid, = args
            node = None
            type_ = 'info'
        elif len(args) == 2:
            jid, node = args
            type_ = 'info'
        else:
            jid, node, type_ = args
        try:
            if type_ == 'info':
                iq = await self.core.xmpp.plugin['xep_0030'].get_info(
                    jid=JID(jid), node=node, cached=False
                )
                self.on_info(iq)
            elif type_ == 'items':
                iq = await self.core.xmpp.plugin['xep_0030'].get_items(
                    jid=JID(jid), node=node
                )
                self.on_items(iq)
        except InvalidJID as e:
            self.api.information('Invalid JID “%s”: %s' % (jid, e), 'Error')
        except IqError as e:
            self.api.information('Received iq error while querying “%s”: %s' % (jid, e), 'Error')
        except IqTimeout:
            self.api.information('Received no reply querying “%s”…' % jid, 'Error')



"""
This plugin makes you have a random nick when joining a chatroom.

Usage
-----

To have a random nick, just join a room with “RANDOM” as your nick. It will
automatically be changed to something random, for example: ::

    /join coucou@conference.example.com/RANDOM

"""

from poezio.plugin import BasePlugin
from slixmpp import Presence
from random import choice


class Plugin(BasePlugin):
    def init(self) -> None:
        self.api.add_event_handler('joining_muc', self.change_nick_to_random)
        self.api.add_event_handler('changing_nick', self.change_nick_to_random)

    def change_nick_to_random(self, presence: Presence) -> None:
        to = presence["to"]
        if to.resource == 'RANDOM':
            to.resource = gen_nick(3)
            presence["to"] = to


s = ["i", "ou", "ou", "on", "a", "o", "u", "i"]
c = [
    "b", "c", "d", "f", "g", "h", "j", "k", "m", "l", "n", "p", "r", "s", "t",
    "v", "z"
]


def gen_nick(size: int) -> str:
    res = ''
    for _ in range(size):
        res += '%s%s' % (choice(c), choice(s))
    return res



from __future__ import annotations

from datetime import datetime
from math import ceil, log10
from typing import Optional, Any, Callable

from slixmpp import JID

from poezio import poopt
from poezio.theming import dump_tuple, get_theme
from poezio.ui.funcs import truncate_nick
from poezio.user import User
from poezio.theming import TextStyle


class BaseMessage:
    """Base class for all ui-related messages"""
    __slots__ = ('txt', 'time', 'identifier')

    txt: str
    identifier: str
    time: datetime

    def __init__(self, txt: str, identifier: str = '', time: Optional[datetime] = None):
        self.txt = txt
        self.identifier = identifier
        if time is not None:
            self.time = time
        else:
            self.time = datetime.now()

    def compute_offset(self, with_timestamps: bool, nick_size: int) -> int:
        """Compute the offset of the message"""
        theme = get_theme()
        return theme.SHORT_TIME_FORMAT_LENGTH + 1


class EndOfArchive(BaseMessage):
    """Marker added to a buffer when we reach the end of a MAM archive"""


class InfoMessage(BaseMessage):
    """Information message"""
    def __init__(self, txt: str, identifier: str = '', time: Optional[datetime] = None):
        txt = ('\x19%s}' % dump_tuple(get_theme().COLOR_INFORMATION_TEXT)) + txt
        super().__init__(txt=txt, identifier=identifier, time=time)


class UIMessage(BaseMessage):
    """Message displayed through poezio UI"""
    __slots__ = ('level', 'color')
    level: str
    color: TextStyle | None

    def __init__(self, txt: str, level: str):
        BaseMessage.__init__(self, txt=txt)
        self.level = level.capitalize()
        colors = get_theme().INFO_COLORS
        self.color = colors.get(level.lower(), colors.get('default', None))

    def compute_offset(self, with_timestamps: bool, nick_size: int) -> int:
        """Compute the x-position at which the message should be printed"""
        offset = 0
        theme = get_theme()
        if with_timestamps:
            offset += 1 + theme.SHORT_TIME_FORMAT_LENGTH
        level = self.level
        if not level:  # not a message, nothing to do afterwards
            return offset
        level = truncate_nick(level, nick_size) or ''
        offset += poopt.wcswidth(level)
        offset += 2
        return offset


class LoggableTrait:
    """Trait for classes of messages that should go through the logger"""
    pass


class PersistentInfoMessage(InfoMessage, LoggableTrait):
    """Information message that will be logged"""
    pass


class MucJoin(PersistentInfoMessage):
    """A join from someone else"""
    pass


class MucLeave(PersistentInfoMessage):
    """A leave from someone else"""
    pass


class MucOwnLeaveMessage(InfoMessage, LoggableTrait):
    """Status message displayed on our room leave/kick/ban"""


class MucOwnJoinMessage(InfoMessage, LoggableTrait):
    """Status message displayed on our room join"""


class XMLLog(BaseMessage):
    """XML Log message"""
    __slots__ = ('incoming')
    incoming: bool

    def __init__(
            self,
            txt: str,
            incoming: bool,
    ):
        BaseMessage.__init__(
            self,
            txt=txt,
        )
        self.incoming = incoming

    def compute_offset(self, with_timestamps: bool, nick_size: int) -> int:
        offset = 0
        theme = get_theme()
        if with_timestamps:
            offset += 1 + theme.SHORT_TIME_FORMAT_LENGTH
        if self.incoming:
            nick = theme.CHAR_XML_IN
        else:
            nick = theme.CHAR_XML_OUT
        nick = truncate_nick(nick, nick_size) or ''
        offset += 1 + len(nick)
        return offset


class StatusMessage(BaseMessage):
    """A dynamically formatted status message"""
    __slots__ = ('format_string', 'format_args')
    format_string: str
    format_args: dict[str, Callable[[], Any]]

    def __init__(self, format_string: str, format_args: dict[str, Callable[[], Any]]) -> None:
        BaseMessage.__init__(
            self,
            txt='',
        )
        self.format_string = format_string
        self.format_args = format_args
        self.rebuild()

    def rebuild(self) -> None:
        """Rebuild the text from available data."""
        real_args = {}
        for key, func in self.format_args.items():
            real_args[key] = func()
        self.txt = self.format_string.format(**real_args)


class Message(BaseMessage, LoggableTrait):
    """The standard text message with a huge collection of possible attributes."""
    __slots__ = ('nick_color', 'nickname', 'user', 'delayed', 'history',
                 'highlight', 'me', 'old_message', 'revisions',
                 'jid', 'stanza_id', 'ack', 'reactions')
    nick_color: TextStyle | None
    nickname: Optional[str]
    user: Optional[User]
    delayed: bool
    history: bool
    highlight: bool
    me: bool
    old_message: Optional[Message]
    revisions: int
    jid: Optional[JID]
    stanza_id: Optional[str]
    ack: int
    reactions: dict[str, set[str]] | None

    def __init__(self,
                 txt: str,
                 nickname: Optional[str],
                 time: Optional[datetime] = None,
                 nick_color: TextStyle | None = None,
                 delayed: bool = False,
                 history: bool = False,
                 user: Optional[User] = None,
                 identifier: Optional[str] = '',
                 highlight: bool = False,
                 old_message: Optional[Message] = None,
                 revisions: int = 0,
                 jid: Optional[JID] = None,
                 stanza_id: Optional[str] = None,
                 ack: int = 0,
                 reactions: dict[str, set[str]] | None = None) -> None:
        """
        Create a new Message object with parameters, check for /me messages,
        and delayed messages
        """
        BaseMessage.__init__(
            self,
            txt=txt.replace('\t', '    ') + '\x19o',
            identifier=identifier or '',
            time=time,
        )
        if txt.startswith('/me '):
            me = True
            txt = '\x19%s}%s\x19o' % (dump_tuple(get_theme().COLOR_ME_MESSAGE),
                                      txt[4:])
        else:
            me = False
        self.txt = txt
        self.delayed = delayed or history
        self.history = history
        self.nickname = nickname
        self.nick_color = nick_color
        self.user = user
        self.highlight = highlight
        self.me = me
        self.old_message = old_message
        self.revisions = revisions
        self.jid = jid
        self.stanza_id = stanza_id
        self.ack = ack
        self.reactions = reactions

    def _other_elems(self) -> str:
        "Helper for the repr_message function"
        acc = []
        fields = list(self.__slots__)
        fields.remove('old_message')
        for field in fields:
            acc.append('%s=%s' % (field, repr(getattr(self, field))))
        return 'Message(%s, %s' % (', '.join(acc), 'old_message=')

    def __repr__(self) -> str:
        """
        repr() for the Message class, for debug purposes, since the default
        repr() is recursive, so it can stack overflow given too many revisions
        of a message
        """
        init = self._other_elems()
        acc = [init]
        next_message = self.old_message
        rev = 1
        while next_message is not None:
            acc.append(next_message._other_elems())
            next_message = next_message.old_message
            rev += 1
        acc.append('None')
        while rev:
            acc.append(')')
            rev -= 1
        return ''.join(acc)

    def compute_offset(self, with_timestamps: bool, nick_size: int) -> int:
        """Compute the x-position at which the message should be printed"""
        offset = 0
        theme = get_theme()
        if with_timestamps:
            if self.history:
                offset += 1 + theme.LONG_TIME_FORMAT_LENGTH
            else:
                offset += 1 + theme.SHORT_TIME_FORMAT_LENGTH

        if not self.nickname:  # not a message, nothing to do afterwards
            return offset

        nick = truncate_nick(self.nickname, nick_size) or ''
        offset += poopt.wcswidth(nick)
        if self.ack:
            theme = get_theme()
            if self.ack > 0:
                offset += poopt.wcswidth(theme.CHAR_ACK_RECEIVED) + 1
            else:
                offset += poopt.wcswidth(theme.CHAR_NACK) + 1
        if self.me:
            offset += 3
        else:
            offset += 2
        if self.revisions:
            offset += ceil(log10(self.revisions + 1))
        return offset


class DeletedMessage(PersistentInfoMessage):
    """Placeholder for a deleted message"""
    __slots__ = ('old_message',)
    old_message: Message | None

    def __init__(self, txt: str, identifier: str = '', time: datetime | None = None,
                 old_message: Message | None = None) -> None:
        super().__init__(txt=txt, identifier=identifier, time=time)
        self.old_message = old_message



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2018 Maxime “pep” Buquet
# Copyright © 2019 Madhur Garg
#
# Distributed under terms of the GPL-3.0+ license. See the COPYING file.

"""
    Search provided string in the buffer and return all results on the screen
"""

import re
from typing import Optional
from datetime import datetime

from poezio.plugin import BasePlugin
from poezio import tabs
from poezio.text_buffer import TextBuffer
from poezio.ui.types import Message as PMessage, InfoMessage


def add_line(
        text_buffer: TextBuffer,
        text: str,
        datetime: Optional[datetime] = None,
) -> None:
    """Adds a textual entry in the TextBuffer"""
    text_buffer.add_message(InfoMessage(text, time=datetime))


class Plugin(BasePlugin):
    """Lastlog Plugin"""

    def init(self) -> None:
        for tab in tabs.DynamicConversationTab, tabs.StaticConversationTab, tabs.PrivateTab, tabs.MucTab:
            self.api.add_tab_command(
                tab,
                'lastlog',
                self.command_lastlog,
                usage='<keyword>',
                help=(
                    'Search <keyword> in the buffer and returns results'
                    'on the screen'
                ),
            )

    def command_lastlog(self, input_: str) -> None:
        """Define lastlog command"""

        current_tab = self.api.current_tab()
        if not isinstance(current_tab, tabs.ChatTab):
            return
        text_buffer = current_tab._text_buffer
        search_re = re.compile(input_, re.I)

        res = []
        add_line(text_buffer, "Lastlog:")
        for message in text_buffer.messages:
            if isinstance(message, PMessage) and \
               search_re.search(message.txt) is not None:
                res.append(message)
                add_line(text_buffer, "%s> %s" % (message.nickname, message.txt), message.time)
        add_line(text_buffer, "End of Lastlog")
        current_tab.text_win.pos = 0
        current_tab.core.refresh_window()



"""
This plugin adds a command (that can be bound to a key) that adds a random
number of dots in the input, making you look depressed, or overly thinking...

Installation
------------
Load the plugin.::

    /load pointpoint

Then use the command: ::

    /pointpoint

But since the goal is to be able to add the dots while typing a message,
entering a command is not really useful. To be useful, this plugin needs to
be used through a bound key, for example like this: ::

    /bind M-. _exc_pointpoint

You just need to press Alt+. and this will insert dots in your message.

Command
-------

.. glossary::

    /pointpoint
        **Usage:** ``/pointpoint``

        …


"""

from poezio.plugin import BasePlugin
from random import randrange


class Plugin(BasePlugin):
    def init(self) -> None:
        self.api.add_command(
            'pointpoint',
            self.command_pointpoint,
            help='Insert a random number of dots in the input')

    def command_pointpoint(self, args: str) -> None:
        for _ in range(randrange(8, 25)):
            self.api.current_tab().input.do_command(".")



# poezio emoji_ascii plugin
#
# Will translate received Emoji to :emoji: for better display on text terminals,
# and outgoing :emoji: into Emoji on the wire.
#
# Requires emojis.json.gz (MIT licensed) from:
#
#     git clone https://github.com/vdurmont/emoji-java
#     gzip -9 < ./src/main/resources/emojis.json > poezio/plugins/emojis.json.gz

# TODOs:
# 1. it messes up your log files (doesn't log original message, logs mutilated :emoji: instead)
# 2. Doesn't work on outgoing direct messages
# 3. Doesn't detect pastes, corrupts jabber:x:foobar
# 4. no auto-completion of emoji aliases
# 5. coloring of converted Emojis to be able to differentiate them from incoming ASCII

import gzip
import json
import os
import re

from poezio.plugin import BasePlugin
from slixmpp import Message
from typing import Any


class Plugin(BasePlugin):
    emoji_to_ascii: dict[str, str] = {}
    ascii_to_emoji: dict[str, str] = {}
    emoji_pattern: re.Pattern[str]
    alias_pattern: re.Pattern[str]

    def init(self) -> None:
        emoji_map_file_name = os.path.abspath(os.path.dirname(__file__) + '/emojis.json.gz')
        emoji_map_data = gzip.open(emoji_map_file_name, 'r').read().decode('utf-8')
        emoji_map = json.loads(emoji_map_data)
        for e in emoji_map:
            self.emoji_to_ascii[e['emoji']] = ':%s:' % e['aliases'][0]
            for alias in e['aliases']:
                # work around :iq: and similar country code misdetection
                flag = re.match('^[a-z][a-z]$', alias) and "flag" in e["tags"]
                if not flag:
                    self.ascii_to_emoji[':%s:' % alias] = e['emoji']
        self.emoji_pattern = re.compile('|'.join(self.emoji_to_ascii.keys()).replace('*', r'\*'))
        self.alias_pattern = re.compile('|'.join(self.ascii_to_emoji.keys()).replace('+', r'\+'))

        self.api.add_event_handler('muc_msg', self.emoji2alias)
        self.api.add_event_handler('conversation_msg', self.emoji2alias)
        self.api.add_event_handler('private_msg', self.emoji2alias)

        self.api.add_event_handler('muc_say', self.alias2emoji)
        self.api.add_event_handler('private_say', self.alias2emoji)
        self.api.add_event_handler('conversation_say', self.alias2emoji)

    def emoji2alias(self, msg: Message, tab: Any) -> None:
        msg['body'] = self.emoji_pattern.sub(lambda m: self.emoji_to_ascii[m.group()], msg['body'])

    def alias2emoji(self, msg: Message, tab: Any) -> None:
        msg['body'] = self.alias_pattern.sub(lambda m: self.ascii_to_emoji[m.group()], msg['body'])



"""
Defines the roster sorting methods used in roster.py
(for contacts/groups)
"""

from typing import TYPE_CHECKING, Callable


if TYPE_CHECKING:
    from poezio.contact import Contact
    from poezio.roster import RosterGroup


########################### Contacts sorting ############################

PRESENCE_PRIORITY = {
    'unavailable': 5,
    'xa': 4,
    'away': 3,
    'dnd': 2,
    '': 1,
    'available': 1
}


def sort_jid(contact: 'Contact') -> str:
    """Sort by contact JID"""
    return contact.bare_jid.bare


def sort_show(contact: 'Contact') -> int:
    """Sort by show (from high availability to low)"""
    res = contact.get_highest_priority_resource()
    if not res:
        return 5
    show = res.presence
    if show not in PRESENCE_PRIORITY:
        return 0
    return PRESENCE_PRIORITY[show]


def sort_resource_nb(contact: 'Contact') -> int:
    """Sort by number of connected resources"""
    return -len(contact)


def sort_name(contact: 'Contact') -> str:
    """Sort by name (case insensitive)"""
    return contact.name.lower() or contact.bare_jid.bare


def sort_sname(contact: 'Contact') -> str:
    """Sort by name (case sensitive)"""
    return contact.name or contact.bare_jid.bare


def sort_online(contact: 'Contact') -> int:
    """Sort by connected/disconnected"""
    result = sort_show(contact)
    return 0 if result < 5 else 1


SORTING_METHODS: dict[str, Callable[['Contact'],  int | str]] = {
    'jid': sort_jid,
    'sname': sort_sname,
    'show': sort_show,
    'resource': sort_resource_nb,
    'name': sort_name,
    'online': sort_online,
}

######################## Roster Groups sorting ##########################


def sort_group_name(group: 'RosterGroup') -> str:
    """Sort by name (case insensitive)"""
    return group.name.lower()


def sort_group_sname(group: 'RosterGroup') -> str:
    """Sort by name (case-sensitive)"""
    return group.name


def sort_group_folded(group: 'RosterGroup') -> bool:
    """Sort by folded/unfolded"""
    return group.folded


def sort_group_connected(group: 'RosterGroup') -> int:
    """Sort by number of connected contacts"""
    return -group.get_nb_connected_contacts()


def sort_group_size(group: 'RosterGroup') -> int:
    """Sort by group size"""
    return -len(group)


def sort_group_none(group: 'RosterGroup') -> int:
    """Put the none group at the end, if any"""
    return 0 if group.name != 'none' else 1


GROUP_SORTING_METHODS: dict[str, Callable[['RosterGroup'], int | str]] = {
    'name': sort_group_name,
    'fold': sort_group_folded,
    'connected': sort_group_connected,
    'size': sort_group_size,
    'none': sort_group_none,
    'sname': sort_group_sname,
}



"""
This plugin enables rich presence events, such as mood, activity, gaming or tune.

.. versionadded:: 0.14
    This plugin was previously provided in the poezio core features.

Command
-------
.. glossary::

    /activity
        **Usage:** ``/activity [<general> [specific] [comment]]``

        Send your current activity to your contacts (use the completion to cycle
        through all the general and specific possible activities).

        Nothing means "stop broadcasting an activity".

    /mood
        **Usage:** ``/mood [<mood> [comment]]``
        Send your current mood to your contacts (use the completion to cycle
        through all the possible moods).

        Nothing means "stop broadcasting a mood".

    /gaming
        **Usage:** ``/gaming [<game name> [server address]]``

        Send your current gaming activity to your contacts.

        Nothing means "stop broadcasting a gaming activity".


Configuration
-------------

.. glossary::

    display_gaming_notifications

        **Default value:** ``true``

        If set to true, notifications about the games your contacts are playing
        will be displayed in the info buffer as 'Gaming' messages.

    display_tune_notifications

        **Default value:** ``true``

        If set to true, notifications about the music your contacts listen to
        will be displayed in the info buffer as 'Tune' messages.

    display_mood_notifications

        **Default value:** ``true``

        If set to true, notifications about the mood of your contacts
        will be displayed in the info buffer as 'Mood' messages.

    display_activity_notifications

        **Default value:** ``true``

        If set to true, notifications about the current activity of your contacts
        will be displayed in the info buffer as 'Activity' messages.

    enable_user_activity

        **Default value:** ``true``

        Set this to ``false`` if you don’t want to receive the activity of your contacts.

    enable_user_gaming

        **Default value:** ``true``

        Set this to ``false`` if you don’t want to receive the gaming activity of your contacts.

    enable_user_mood

        **Default value:** ``true``

        Set this to ``false`` if you don’t want to receive the mood of your contacts.

    enable_user_tune

        **Default value:** ``true``

        If this is set to ``false``, you will no longer be subscribed to tune events,
        and the :term:`display_tune_notifications` option will be ignored.


"""
import asyncio
from typing import TYPE_CHECKING

from slixmpp import Message
from poezio.decorators import command_args_parser
from poezio.plugin import BasePlugin
from poezio.roster import roster
from poezio.core.structs import Completion
from poezio import common

if TYPE_CHECKING:
    from poezio.windows.inputs import Input


class Plugin(BasePlugin):

    default_config = {
        'user_extras': {
            'display_gaming_notifications': True,
            'display_mood_notifications': True,
            'display_activity_notifications': True,
            'display_tune_notifications': True,
            'enable_user_activity': True,
            'enable_user_gaming': True,
            'enable_user_mood': True,
            'enable_user_tune': True,
        }
    }

    def init(self) -> None:
        for plugin in {'xep_0196', 'xep_0108', 'xep_0107', 'xep_0118'}:
            self.core.xmpp.register_plugin(plugin)
        self.api.add_command(
            'activity',
            self.command_activity,
            usage='[<general> [specific] [text]]',
            help='Send your current activity to your contacts '
            '(use the completion). Nothing means '
            '"stop broadcasting an activity".',
            short='Send your activity.',
            completion=self.comp_activity
        )
        self.api.add_command(
            'mood',
            self.command_mood,
            usage='[<mood> [text]]',
            help='Send your current mood to your contacts '
            '(use the completion). Nothing means '
            '"stop broadcasting a mood".',
            short='Send your mood.',
            completion=self.comp_mood,
        )
        self.api.add_command(
            'gaming',
            self.command_gaming,
            usage='[<game name> [server address]]',
            help='Send your current gaming activity to '
            'your contacts. Nothing means "stop '
            'broadcasting a gaming activity".',
            short='Send your gaming activity.',
            completion=None
        )
        handlers = [
            ('user_mood_publish', self.on_mood_event),
            ('user_tune_publish', self.on_tune_event),
            ('user_gaming_publish', self.on_gaming_event),
            ('user_activity_publish', self.on_activity_event),
        ]
        for name, handler in handlers:
            self.core.xmpp.add_event_handler(name, handler)

    def cleanup(self) -> None:
        handlers = [
            ('user_mood_publish', self.on_mood_event),
            ('user_tune_publish', self.on_tune_event),
            ('user_gaming_publish', self.on_gaming_event),
            ('user_activity_publish', self.on_activity_event),
        ]
        for name, handler in handlers:
            self.core.xmpp.del_event_handler(name, handler)
        self.core.register_task(self._stop())

    async def _stop(self) -> None:
        await asyncio.gather(
            self.core.xmpp.plugin['xep_0108'].stop(),
            self.core.xmpp.plugin['xep_0107'].stop(),
            self.core.xmpp.plugin['xep_0196'].stop(),
        )

    @command_args_parser.quoted(0, 2)
    async def command_mood(self, args: list[str]) -> None:
        """
        /mood [<mood> [text]]
        """
        if not args:
            await self.core.xmpp.plugin['xep_0107'].stop()
            return
        mood = args[0]
        if mood not in MOODS:
            self.core.information(
                '%s is not a correct value for a mood.' % mood, 'Error')
            return
        if len(args) == 2:
            text = args[1]
        else:
            text = None
        await self.core.xmpp.plugin['xep_0107'].publish_mood(
            mood, text
        )

    @command_args_parser.quoted(0, 3)
    async def command_activity(self, args: list[str]) -> None:
        """
        /activity [<general> [specific] [text]]
        """
        length = len(args)
        if not length:
            await self.core.xmpp.plugin['xep_0108'].stop()
            return

        general = args[0]
        if general not in ACTIVITIES:
            self.api.information(
                '%s is not a correct value for an activity' % general, 'Error')
            return
        specific = None
        text = None
        if length == 2:
            if args[1] in ACTIVITIES[general]:
                specific = args[1]
            else:
                text = args[1]
        elif length == 3:
            specific = args[1]
            text = args[2]
        if specific and specific not in ACTIVITIES[general]:
            self.core.information(
                '%s is not a correct value '
                'for an activity' % specific, 'Error')
            return
        await self.core.xmpp.plugin['xep_0108'].publish_activity(
            general, specific, text
        )

    @command_args_parser.quoted(0, 2)
    async def command_gaming(self, args: list[str]) -> None:
        """
        /gaming [<game name> [server address]]
        """
        if not args:
            await self.core.xmpp.plugin['xep_0196'].stop()
        else:
            name = args[0]
            if len(args) > 1:
                address = args[1]
            else:
                address = None
            await self.core.xmpp.plugin['xep_0196'].publish_gaming(
                name=name, server_address=address
            )

    def comp_activity(self, the_input: 'Input') -> Completion | None:
        """Completion for /activity"""
        n = the_input.get_argument_position(quoted=True)
        args = common.shell_split(the_input.text)
        if n == 1:
            return Completion(
                the_input.new_completion,
                sorted(ACTIVITIES.keys()),
                n,
                quotify=True)
        elif n == 2:
            if args[1] in ACTIVITIES:
                activities = list(ACTIVITIES[args[1]])
                activities.remove('category')
                activities.sort()
                return Completion(the_input.new_completion, activities, n, quotify=True)
        return None

    def comp_mood(self, the_input: 'Input') -> Completion | None:
        """Completion for /mood"""
        n = the_input.get_argument_position(quoted=True)
        if n == 1:
            return Completion(
                the_input.new_completion,
                sorted(MOODS.keys()),
                1,
                quotify=True)
        return None

    def on_gaming_event(self, message: Message) -> None:
        """
        Called when a pep notification for user gaming
        is received
        """
        contact = roster[message['from'].bare]
        if not contact:
            return
        item = message['pubsub_event']['items']['item']
        old_gaming = contact.rich_presence['gaming']
        xml_node = item.xml.find('{urn:xmpp:gaming:0}game')
        # list(xml_node) checks whether there are children or not.
        if xml_node is not None and list(xml_node):
            item = item['gaming']
            # only name and server_address are used for now
            contact.rich_presence['gaming'] = {
                'character_name': item['character_name'],
                'character_profile': item['character_profile'],
                'name': item['name'],
                'level': item['level'],
                'uri': item['uri'],
                'server_name': item['server_name'],
                'server_address': item['server_address'],
            }
        else:
            contact.rich_presence['gaming'] = {}

        if old_gaming != contact.rich_presence['gaming'] and self.config.get(
                'display_gaming_notifications'):
            if contact.rich_presence['gaming']:
                self.core.information(
                    '%s is playing %s' % (contact.bare_jid,
                                          common.format_gaming_string(
                                              contact.rich_presence['gaming'])), 'Gaming')
            else:
                self.core.information(f'{contact.bare_jid} stopped playing.',
                                      'Gaming')

    def on_mood_event(self, message: Message) -> None:
        """
        Called when a pep notification for a user mood
        is received.
        """
        contact = roster[message['from'].bare]
        if not contact:
            return
        item = message['pubsub_event']['items']['item']
        old_mood = contact.rich_presence.get('mood')
        plugin = item.get_plugin('mood', check=True)
        if plugin:
            mood = item['mood']['value']
        else:
            mood = ''
        if mood:
            mood = MOODS.get(mood, mood)
            text = item['mood']['text']
            if text:
                mood = '%s (%s)' % (mood, text)
            contact.rich_presence['mood'] = mood
        else:
            contact.rich_presence['mood'] = ''

        if old_mood != contact.rich_presence['mood'] and self.config.get(
                'display_mood_notifications'):
            if contact.rich_presence['mood']:
                self.core.information(
                    f'Mood from {contact.bare_jid}: {contact.rich_presence["mood"]}',
                    'Mood')
            else:
                self.core.information(
                    f'{contact.bare_jid} stopped having their mood.', 'Mood')

    def on_activity_event(self, message: Message) -> None:
        """
        Called when a pep notification for a user activity
        is received.
        """
        contact = roster[message['from'].bare]
        if not contact:
            return
        item = message['pubsub_event']['items']['item']
        old_activity = contact.rich_presence['activity']
        xml_node = item.xml.find('{http://jabber.org/protocol/activity}activity')
        # list(xml_node) checks whether there are children or not.
        if xml_node is not None and list(xml_node):
            try:
                activity = item['activity']['value']
            except ValueError:
                return
            if activity[0]:
                general = ACTIVITIES.get(activity[0])
                if general is None:
                    return
                s = general['category']
                if activity[1]:
                    s = s + '/' + general.get(activity[1], 'other')
                text = item['activity']['text']
                if text:
                    s = '%s (%s)' % (s, text)
                contact.rich_presence['activity'] = s
            else:
                contact.rich_presence['activity'] = ''
        else:
            contact.rich_presence['activity'] = ''

        if old_activity != contact.rich_presence['activity'] and self.config.get(
                'display_activity_notifications'):
            if contact.rich_presence['activity']:
                self.core.information(
                    f'Activity from {contact.bare_jid}: '
                    f'{contact.rich_presence["activity"]}', 'Activity')
            else:
                self.core.information(
                    f'{contact.bare_jid} stopped doing their activity.',
                    'Activity')

    def on_tune_event(self, message: Message) -> None:
        """
        Called when a pep notification for a user tune
        is received
        """
        contact = roster[message['from'].bare]
        if not contact:
            return
        roster.modified()
        item = message['pubsub_event']['items']['item']
        old_tune = contact.rich_presence['tune']
        xml_node = item.xml.find('{http://jabber.org/protocol/tune}tune')
        # list(xml_node) checks whether there are children or not.
        if xml_node is not None and list(xml_node):
            item = item['tune']
            contact.rich_presence['tune'] = {
                'artist': item['artist'],
                'length': item['length'],
                'rating': item['rating'],
                'source': item['source'],
                'title': item['title'],
                'track': item['track'],
                'uri': item['uri']
            }
        else:
            contact.rich_presence['tune'] = {}

        if old_tune != contact.rich_presence['tune'] and self.config.get(
                'display_tune_notifications'):
            if contact.rich_presence['tune']:
                self.core.information(
                    f'Tune from {message["from"].bare}: '
                    f'{common.format_tune_string(contact.rich_presence["tune"])}',
                    'Tune',
                )
            else:
                self.core.information(
                    f'{contact.bare_jid} stopped listening to music.', 'Tune')


# Collection of mappings for PEP moods/activities
# extracted directly from the XEP

MOODS: dict[str, str] = {
    'afraid': 'Afraid',
    'amazed': 'Amazed',
    'angry': 'Angry',
    'amorous': 'Amorous',
    'annoyed': 'Annoyed',
    'anxious': 'Anxious',
    'aroused': 'Aroused',
    'ashamed': 'Ashamed',
    'bored': 'Bored',
    'brave': 'Brave',
    'calm': 'Calm',
    'cautious': 'Cautious',
    'cold': 'Cold',
    'confident': 'Confident',
    'confused': 'Confused',
    'contemplative': 'Contemplative',
    'contented': 'Contented',
    'cranky': 'Cranky',
    'crazy': 'Crazy',
    'creative': 'Creative',
    'curious': 'Curious',
    'dejected': 'Dejected',
    'depressed': 'Depressed',
    'disappointed': 'Disappointed',
    'disgusted': 'Disgusted',
    'dismayed': 'Dismayed',
    'distracted': 'Distracted',
    'embarrassed': 'Embarrassed',
    'envious': 'Envious',
    'excited': 'Excited',
    'flirtatious': 'Flirtatious',
    'frustrated': 'Frustrated',
    'grumpy': 'Grumpy',
    'guilty': 'Guilty',
    'happy': 'Happy',
    'hopeful': 'Hopeful',
    'hot': 'Hot',
    'humbled': 'Humbled',
    'humiliated': 'Humiliated',
    'hungry': 'Hungry',
    'hurt': 'Hurt',
    'impressed': 'Impressed',
    'in_awe': 'In awe',
    'in_love': 'In love',
    'indignant': 'Indignant',
    'interested': 'Interested',
    'intoxicated': 'Intoxicated',
    'invincible': 'Invincible',
    'jealous': 'Jealous',
    'lonely': 'Lonely',
    'lucky': 'Lucky',
    'mean': 'Mean',
    'moody': 'Moody',
    'nervous': 'Nervous',
    'neutral': 'Neutral',
    'offended': 'Offended',
    'outraged': 'Outraged',
    'playful': 'Playful',
    'proud': 'Proud',
    'relaxed': 'Relaxed',
    'relieved': 'Relieved',
    'remorseful': 'Remorseful',
    'restless': 'Restless',
    'sad': 'Sad',
    'sarcastic': 'Sarcastic',
    'serious': 'Serious',
    'shocked': 'Shocked',
    'shy': 'Shy',
    'sick': 'Sick',
    'sleepy': 'Sleepy',
    'spontaneous': 'Spontaneous',
    'stressed': 'Stressed',
    'strong': 'Strong',
    'surprised': 'Surprised',
    'thankful': 'Thankful',
    'thirsty': 'Thirsty',
    'tired': 'Tired',
    'undefined': 'Undefined',
    'weak': 'Weak',
    'worried': 'Worried'
}

ACTIVITIES: dict[str, dict[str, str]] = {
    'doing_chores': {
        'category': 'Doing_chores',
        'buying_groceries': 'Buying groceries',
        'cleaning': 'Cleaning',
        'cooking': 'Cooking',
        'doing_maintenance': 'Doing maintenance',
        'doing_the_dishes': 'Doing the dishes',
        'doing_the_laundry': 'Doing the laundry',
        'gardening': 'Gardening',
        'running_an_errand': 'Running an errand',
        'walking_the_dog': 'Walking the dog',
        'other': 'Other',
    },
    'drinking': {
        'category': 'Drinking',
        'having_a_beer': 'Having a beer',
        'having_coffee': 'Having coffee',
        'having_tea': 'Having tea',
        'other': 'Other',
    },
    'eating': {
        'category': 'Eating',
        'having_breakfast': 'Having breakfast',
        'having_a_snack': 'Having a snack',
        'having_dinner': 'Having dinner',
        'having_lunch': 'Having lunch',
        'other': 'Other',
    },
    'exercising': {
        'category': 'Exercising',
        'cycling': 'Cycling',
        'dancing': 'Dancing',
        'hiking': 'Hiking',
        'jogging': 'Jogging',
        'playing_sports': 'Playing sports',
        'running': 'Running',
        'skiing': 'Skiing',
        'swimming': 'Swimming',
        'working_out': 'Working out',
        'other': 'Other',
    },
    'grooming': {
        'category': 'Grooming',
        'at_the_spa': 'At the spa',
        'brushing_teeth': 'Brushing teeth',
        'getting_a_haircut': 'Getting a haircut',
        'shaving': 'Shaving',
        'taking_a_bath': 'Taking a bath',
        'taking_a_shower': 'Taking a shower',
        'other': 'Other',
    },
    'having_appointment': {
        'category': 'Having appointment',
        'other': 'Other',
    },
    'inactive': {
        'category': 'Inactive',
        'day_off': 'Day_off',
        'hanging_out': 'Hanging out',
        'hiding': 'Hiding',
        'on_vacation': 'On vacation',
        'praying': 'Praying',
        'scheduled_holiday': 'Scheduled holiday',
        'sleeping': 'Sleeping',
        'thinking': 'Thinking',
        'other': 'Other',
    },
    'relaxing': {
        'category': 'Relaxing',
        'fishing': 'Fishing',
        'gaming': 'Gaming',
        'going_out': 'Going out',
        'partying': 'Partying',
        'reading': 'Reading',
        'rehearsing': 'Rehearsing',
        'shopping': 'Shopping',
        'smoking': 'Smoking',
        'socializing': 'Socializing',
        'sunbathing': 'Sunbathing',
        'watching_a_movie': 'Watching a movie',
        'watching_tv': 'Watching tv',
        'other': 'Other',
    },
    'talking': {
        'category': 'Talking',
        'in_real_life': 'In real life',
        'on_the_phone': 'On the phone',
        'on_video_phone': 'On video phone',
        'other': 'Other',
    },
    'traveling': {
        'category': 'Traveling',
        'commuting': 'Commuting',
        'driving': 'Driving',
        'in_a_car': 'In a car',
        'on_a_bus': 'On a bus',
        'on_a_plane': 'On a plane',
        'on_a_train': 'On a train',
        'on_a_trip': 'On a trip',
        'walking': 'Walking',
        'cycling': 'Cycling',
        'other': 'Other',
    },
    'undefined': {
        'category': 'Undefined',
        'other': 'Other',
    },
    'working': {
        'category': 'Working',
        'coding': 'Coding',
        'in_a_meeting': 'In a meeting',
        'writing': 'Writing',
        'studying': 'Studying',
        'other': 'Other',
    }
}



"""
Text inputs.
"""

import curses
import logging
import string
from typing import List, Callable, Optional, ClassVar, Any

from poezio import keyboard
from poezio import common
from poezio import poopt
from poezio.windows.base_wins import Win
from poezio.ui.consts import FORMAT_CHARS
from poezio.ui.funcs import find_first_format_char
from poezio.config import config
from poezio.theming import to_curses_attr, TextStyle

log = logging.getLogger(__name__)


def default_on_input(x: str) -> bool:
    return False


class InputLike(Win):
    """
    A parent class providing input facilities but without actually doing stuff.
    """
    text: str

    __slots__ = ('text', )

    def __init__(self) -> None:
        self.text = ''

    def do_command(self, key: str, reset: bool = True, raw: bool = False) -> bool:
        return False

    def on_delete(self) -> None:
        return

    def key_enter(self) -> str | None:
        return ''

    def key_end(self, reset: bool = False) -> bool:
        return False

    def refresh(self) -> None:
        return


class Input(InputLike):
    """
    The simplest Input possible, provides just a way to edit a single line
    of text. It also has a clipboard, common to all Inputs.
    Doesn't have any history.
    It doesn't do anything when enter is pressed either.
    This should be herited for all kinds of Inputs, for example MessageInput
    or the little inputs in dataforms, etc, adding specific features (completion etc)
    It features two kinds of completion, but they have to be called from outside (the Tab),
    passing the list of items that can be used to complete. The completion can be used
    in a very flexible way.
    """
    __slots__ = ('key_func', 'pos', 'view_pos', 'on_input', 'color',
                 'last_completion', 'hit_list')

    text_attributes = 'bou1234567ti'
    clipboard = ''  # A common clipboard for all the inputs, this makes
    on_input: Callable[[str], Any]

    # it easy cut and paste text between various input
    def __init__(self) -> None:
        super().__init__()
        self.key_func: dict[str, Callable[[], bool] | Callable[[], None]] = {
            "KEY_LEFT": self.key_left,
            "KEY_RIGHT": self.key_right,
            "KEY_END": self.key_end,
            "KEY_HOME": self.key_home,
            "KEY_DC": self.key_dc,
            '^D': self.key_dc,
            'M-b': self.jump_word_left,
            "M-[1;5D": self.jump_word_left,
            "kRIT5": self.jump_word_right,
            "kLFT5": self.jump_word_left,
            '^W': self.delete_word,
            'M-d': self.delete_next_word,
            '^K': self.delete_end_of_line,
            '^U': self.delete_beginning_of_line,
            '^Y': self.paste_clipboard,
            '^A': self.key_home,
            '^E': self.key_end,
            'M-f': self.jump_word_right,
            "M-[1;5C": self.jump_word_right,
            "KEY_BACKSPACE": self.key_backspace,
            "M-KEY_BACKSPACE": self.delete_word,
            '^?': self.key_backspace,
            "M-^?": self.delete_word,
            # '^J': self.add_line_break,
        }
        self.pos = 0  # The position of the “cursor” in the text
        # (not only in the view)
        self.view_pos = 0  # The position (in the text) of the
        # first character displayed on the
        # screen
        self.on_input = default_on_input  # callback called on any key pressed
        self.color: TextStyle | None = None  # use this color on addstr
        self.last_completion: str | None = None
        self.hit_list: list[str] = []

    def on_delete(self) -> None:
        """
        Remove all references kept to a tab, so that the tab
        can be garbage collected
        """
        del self.key_func

    def set_color(self, color: TextStyle) -> None:
        self.color = color
        self.rewrite_text()

    def is_empty(self) -> bool:
        if self.text:
            return False
        return True

    def is_cursor_at_end(self) -> bool:
        """
        Whether or not the cursor is at the end of the text.
        """
        assert len(self.text) >= self.pos
        if len(self.text) == self.pos:
            return True
        return False

    def jump_word_left(self) -> bool:
        """
        Move the cursor one word to the left
        """
        if self.pos == 0:
            return True
        separators = string.punctuation + ' ' + '\n'
        while self.pos > 0 and self.text[self.pos - 1] in separators:
            self.key_left()
        while self.pos > 0 and self.text[self.pos - 1] not in separators:
            self.key_left()
        return True

    def jump_word_right(self) -> bool:
        """
        Move the cursor one word to the right
        """
        if self.is_cursor_at_end():
            return True
        separators = string.punctuation + ' ' + '\n'
        while not self.is_cursor_at_end() and self.text[self.pos] in separators:
            self.key_right()
        while not self.is_cursor_at_end() and self.text[self.
                                                        pos] not in separators:
            self.key_right()
        return True

    def delete_word(self) -> bool:
        """
        Delete the word just before the cursor
        """
        separators = string.punctuation + ' ' + '\n'
        while self.pos > 0 and self.text[self.pos - 1] in separators:
            self.key_backspace()
        while self.pos > 0 and self.text[self.pos - 1] not in separators:
            self.key_backspace()
        return True

    def delete_next_word(self) -> bool:
        """
        Delete the word just after the cursor
        """
        separators = string.punctuation + ' ' + '\n'
        while not self.is_cursor_at_end() and self.text[self.pos] in separators:
            self.key_dc()
        while not self.is_cursor_at_end() and self.text[self.
                                                        pos] not in separators:
            self.key_dc()
        return True

    def delete_end_of_line(self) -> bool:
        """
        Cut the text from cursor to the end of line
        """
        if self.is_cursor_at_end():
            return False
        Input.clipboard = self.text[self.pos:]
        self.text = self.text[:self.pos]
        self.key_end()
        return True

    def delete_beginning_of_line(self) -> bool:
        """
        Cut the text from cursor to the beginning of line
        """
        if self.pos == 0:
            return True
        Input.clipboard = self.text[:self.pos]
        self.text = self.text[self.pos:]
        self.key_home()
        return True

    def paste_clipboard(self) -> bool:
        """
        Insert what is in the clipboard at the cursor position
        """
        if not Input.clipboard:
            return True
        for letter in Input.clipboard:
            self.do_command(letter, False)
        self.rewrite_text()
        return True

    def key_dc(self) -> bool:
        """
        delete char just after the cursor
        """
        self.reset_completion()
        if self.is_cursor_at_end():
            return True  # end of line, nothing to delete
        self.text = self.text[:self.pos] + self.text[self.pos + 1:]
        self.rewrite_text()
        return True

    def key_home(self) -> bool:
        """
        Go to the beginning of line
        """
        self.reset_completion()
        self.pos = 0
        self.rewrite_text()
        return True

    def key_end(self, reset: bool = False) -> bool:
        """
        Go to the end of line
        """
        if reset:
            self.reset_completion()
        self.pos = len(self.text)
        assert self.is_cursor_at_end()
        self.rewrite_text()
        return True

    def key_left(self, jump: bool = True, reset: bool = True) -> bool:
        """
        Move the cursor one char to the left
        """
        if reset:
            self.reset_completion()
        if self.pos == 0:
            return True
        self.pos -= 1
        if reset:
            self.rewrite_text()
        return True

    def key_right(self, jump: bool = True, reset: bool = True) -> bool:
        """
        Move the cursor one char to the right
        """
        if reset:
            self.reset_completion()
        if self.is_cursor_at_end():
            return True
        self.pos += 1
        if reset:
            self.rewrite_text()
        return True

    def key_backspace(self, reset: bool = True) -> bool:
        """
        Delete the char just before the cursor
        """
        self.reset_completion()
        if self.pos == 0:
            return False
        self.key_left()
        self.key_dc()
        return True

    def auto_completion(self, word_list: List[str], add_after: str = '',
                        quotify: bool = True) -> bool:
        """
        Complete the input, from a list of words
        if add_after is None, we use the value defined in completion
        plus a space, after the completion. If it's a string, we use it after the
        completion (with no additional space)
        """
        if quotify:
            for i, word in enumerate(word_list[:]):
                word_list[i] = '"' + word + '"'
        self.normal_completion(word_list, add_after)
        return True

    def new_completion(self,
                       word_list: List[str],
                       argument_position: int = -1,
                       add_after: str = '',
                       quotify: bool = True,
                       override: bool = False) -> bool:
        """
        Complete the argument at position ``argument_position`` in the input.
        If ``quotify`` is ``True``, then the completion will operate on block of words
        (e.g. "toto titi") whereas if it is ``False``, it will operate on words (e.g
        "toto", "titi").

        The completions may modify other parts of the input when completing an argument,
        for example removing useless double quotes around single-words, or setting the
        space between each argument to only one space.

        The case where we complete the first argument is special, because we complete
        the command, and we do not want to modify anything else in the input.

        This method is the one that should be used if the command being completed
        has several arguments.
        """
        if argument_position == 0:
            self._new_completion_first(word_list)
        else:
            self._new_completion_args(word_list, argument_position, add_after,
                                      quotify, override)
        self.rewrite_text()
        return True

    def _new_completion_args(self,
                             word_list: List[str],
                             argument_position: int = -1,
                             add_after: str = '',
                             quoted: bool = True,
                             override: bool = False) -> None:
        """
        Case for completing arguments with position ≠ 0
        """
        if quoted:
            words = common.shell_split(self.text)
        else:
            words = self.text.split()
        if argument_position >= len(words):
            current = ''
        else:
            current = words[argument_position]

        if quoted:
            split_words = words[1:]
            words = [words[0]]
            for word in split_words:
                if ' ' in word or '\\' in word:
                    words.append('"' + word + '"')
                else:
                    words.append(word)
        current_l = current.lower()
        if self.last_completion is not None:
            self.hit_list.append(self.hit_list.pop(0))
        else:
            if override:
                hit_list = word_list
            else:
                hit_list = []
                for word in word_list:
                    if word.lower().startswith(current_l):
                        hit_list.append(word)
            if not hit_list:
                return
            self.hit_list = hit_list

        if argument_position >= len(words):
            if quoted and ' ' in self.hit_list[0]:
                words.append('"' + self.hit_list[0] + '"')
            else:
                words.append(self.hit_list[0])
        else:
            if quoted and ' ' in self.hit_list[0]:
                words[argument_position] = '"' + self.hit_list[0] + '"'
            else:
                words[argument_position] = self.hit_list[0]

        new_pos = -1
        for i, word in enumerate(words):
            if argument_position >= i:
                new_pos += len(word) + 1

        self.last_completion = self.hit_list[0]
        self.text = words[0] + ' ' + ' '.join(words[1:])
        self.pos = new_pos

    def _new_completion_first(self, word_list: List[str]) -> None:
        """
        Special case of completing the command itself:
        we don’t want to change anything to the input doing that
        """
        space_pos = self.text.find(' ')
        if space_pos != -1:
            current, follow = self.text[:space_pos], self.text[space_pos:]
        else:
            current, follow = self.text, ''

        if self.last_completion:
            self.hit_list.append(self.hit_list.pop(0))
        else:
            hit_list = []
            for word in word_list:
                if word.lower().startswith(current):
                    hit_list.append(word)
            if not hit_list:
                return
            self.hit_list = hit_list

        self.last_completion = self.hit_list[0]
        self.text = self.hit_list[0] + follow
        self.pos = len(self.hit_list[0])

    def get_argument_position(self, quoted: bool = True) -> int:
        """
        Get the argument number at the current position
        """
        command_stop = self.text.find(' ')
        if command_stop == -1 or self.pos <= command_stop:
            return 0
        text = self.text[command_stop + 1:]
        pos = self.pos - len(self.text) + len(text) - 1
        val = common.find_argument(pos, text, quoted=quoted) + 1
        return val

    def reset_completion(self) -> None:
        """
        Reset the completion list (called on ALL keys except tab)
        """
        self.hit_list = []
        self.last_completion = None

    def normal_completion(self, word_list: List[str], after: str) -> None:
        """
        Normal completion
        """
        pos = self.pos
        if pos < len(self.text) and after.endswith(' ') and self.text[pos] in ' \n':
            after = after[:
                          -1]  # remove the last space if we are already on a space
        if not self.last_completion:
            space_before_cursor = self.text.rfind(' ', 0, pos)
            line_before_cursor = self.text.rfind('\n', 0, pos)
            if line_before_cursor > space_before_cursor:
                space_before_cursor = line_before_cursor
            if space_before_cursor != -1:
                begin = self.text[space_before_cursor + 1:pos]
            else:
                begin = self.text[:pos]
            hit_list = []  # list of matching hits
            for word in word_list:
                if word.lower().startswith(begin.lower()):
                    hit_list.append(word)
                elif word.startswith('"') and word.lower()[1:].startswith(
                        begin.lower()):
                    hit_list.append(word)
            if len(hit_list) == 0:
                return
            self.hit_list = hit_list
            end = len(begin)
        else:
            begin = self.last_completion
            end = len(begin) + len(after)
            self.hit_list.append(self.hit_list.pop(0))  # rotate list

        self.text = self.text[:pos - end] + self.text[pos:]
        pos -= end
        hit = self.hit_list[0]  # take the first hit
        self.text = self.text[:pos] + hit + after + self.text[pos:]
        for _ in range(end):
            try:
                self.key_left(reset=False)
            except Exception as exc:
                log.debug('Key: unable to go left: %s', exc)
        for _ in range(len(hit) + len(after)):
            self.key_right(reset=False)

        self.rewrite_text()
        self.last_completion = hit

    def do_command(self, key: str, reset: bool = True, raw: bool = False) -> bool:
        if key in self.key_func:
            res = self.key_func[key]()
            if not raw and self.on_input is not default_on_input:
                self.on_input(self.get_text())
            return bool(res)
        if not raw and (not key or len(key) > 1):
            return False  # ignore non-handled keyboard shortcuts
        if reset:
            self.reset_completion()
        # Insert the char at the cursor position
        self.text = self.text[:self.pos] + key + self.text[self.pos:]
        self.pos += len(key)
        if reset:
            self.rewrite_text()
        if self.on_input is not default_on_input:
            self.on_input(self.get_text())

        return True

    def add_line_break(self) -> None:
        """
        Add a (real) \n to the line
        """
        self.do_command('\n')

    def get_text(self) -> str:
        """
        Return the text entered so far
        """
        return self.text

    def _addstr_colored_lite(self, text: str, y: Optional[int] = None,
                             x: Optional[int] = None) -> None:
        """
        Just like addstr_colored, with the single-char attributes
        (\x0E to \x19 instead of \x19 + attr). We do not use any }
        char in this version
        """
        chars = FORMAT_CHARS + '\n'
        if y is not None and x is not None:
            self.move(y, x)
        format_char = find_first_format_char(text, chars)
        attr_italic = curses.A_ITALIC if hasattr(
            curses, 'A_ITALIC') else curses.A_REVERSE
        while format_char != -1:
            if text[format_char] == '\n':
                attr_char = '|'
            else:
                attr_char = self.text_attributes[FORMAT_CHARS.index(
                    text[format_char])]
            self.addstr(text[:format_char])
            self.addstr(attr_char, curses.A_REVERSE)
            text = text[format_char + 1:]
            if attr_char == 'o':
                self._win.attrset(0)
            elif attr_char == 'u':
                self._win.attron(curses.A_UNDERLINE)
            elif attr_char == 'b':
                self._win.attron(curses.A_BOLD)
            elif attr_char == 'i':
                self._win.attron(attr_italic)
            elif attr_char in string.digits and attr_char != '':
                self._win.attron(to_curses_attr((int(attr_char), -1)))
            format_char = find_first_format_char(text, chars)
        self.addstr(text)

    def rewrite_text(self) -> None:
        """
        Refresh the line onscreen, but first, always adjust the
        view_pos.  Also, each FORMAT_CHAR+attr_char count only take
        one screen column (this is done in _addstr_colored_lite), we
        have to do some special calculations to find the correct
        length of text to display, and the position of the cursor.
        """
        self.adjust_view_pos()
        text = self.text
        self._win.erase()
        if self.color:
            self._win.attron(to_curses_attr(self.color))
        displayed_text = text[self.view_pos:self.view_pos + self.width -
                              1].replace('\t', '\x18')
        self._win.attrset(0)
        self._addstr_colored_lite(displayed_text)
        # Fill the rest of the line with the input color
        if self.color:
            (_, x) = self._win.getyx()
            size = self.width - x
            self.addnstr(' ' * size, size, to_curses_attr(self.color))
        self.addstr(0, poopt.wcswidth(
            displayed_text[:self.pos - self.view_pos]), '')
        if self.color:
            self._win.attroff(to_curses_attr(self.color))
        curses.curs_set(1)
        self._refresh()

    def adjust_view_pos(self) -> None:
        """
        Adjust the position of the View, if needed (for example if the
        cursor moved and would now be out of the view, we adapt the
        view_pos so that we can always see our cursor)
        """
        # start of the input
        if self.pos == 0:
            self.view_pos = 0
            return
        # cursor outside of the screen (left)
        if self.pos <= self.view_pos:
            self.view_pos = self.pos - max(1 * self.width // 3, 1)
        # cursor outside of the screen (right)
        elif self.pos >= self.view_pos + self.width - 1:
            self.view_pos = self.pos - max(2 * self.width // 3, 2)

        if self.view_pos < 0:
            self.view_pos = 0

        # text small enough to fit inside the window entirely:
        # remove scrolling if present
        if poopt.wcswidth(self.text) < self.width:
            self.view_pos = 0

    def refresh(self) -> None:
        log.debug('Refresh: %s', self.__class__.__name__)
        self.rewrite_text()

    def clear_text(self) -> None:
        self.text = ''
        self.pos = 0
        self.rewrite_text()

    def key_enter(self) -> Optional[str]:
        txt = self.get_text()
        self.clear_text()
        return txt


class HistoryInput(Input):
    """
    An input with colors and stuff, plus an history
    ^R allows to search inside the history (as in a shell)
    """
    __slots__ = ('help_message', 'histo_pos', 'current_completed', 'search',
                 'history')

    global_history: ClassVar[List[str]] = []

    def __init__(self) -> None:
        Input.__init__(self)
        self.help_message = ''
        self.histo_pos = -1
        self.current_completed = ''
        self.key_func['^R'] = self.toggle_search
        self.search = False
        if config.getbool('separate_history'):
            self.history: List[str] = []
        else:
            self.history = self.__class__.global_history

    def toggle_search(self) -> None:
        if self.help_message:
            return
        self.search = not self.search
        self.refresh()

    def update_completed(self) -> None:
        """
        Find a match for the current text
        """
        if not self.text:
            return
        for i in self.history:
            if self.text in i:
                self.current_completed = i
                return
        self.current_completed = ''

    def history_enter(self) -> bool:
        """
        Enter was pressed, set the text to the
        current completion and disable history
        search
        """
        if self.search:
            self.search = False
            if self.current_completed:
                self.text = self.current_completed
                self.current_completed = ''
            self.refresh()
            return True
        self.refresh()
        return False

    def key_up(self) -> bool:
        """
        Get the previous line in the history
        """
        self.reset_completion()
        if self.histo_pos == -1 and self.get_text():
            if not self.history or self.history[0] != self.get_text():
                # add the message to history, we do not want to lose it
                self.history.insert(0, self.get_text())
                self.histo_pos += 1
        if self.histo_pos < len(self.history) - 1:
            self.histo_pos += 1
            self.text = self.history[self.histo_pos]
        self.key_end()
        return True

    def key_down(self) -> bool:
        """
        Get the next line in the history
        """
        self.reset_completion()
        if self.histo_pos > 0:
            self.histo_pos -= 1
            self.text = self.history[self.histo_pos]
        elif self.histo_pos <= 0 and self.get_text():
            if not self.history or self.history[0] != self.get_text():
                # add the message to history, we do not want to lose it
                self.history.insert(0, self.get_text())
            self.text = ''
            self.histo_pos = -1
        self.key_end()
        return True


class MessageInput(HistoryInput):
    """
    The input featuring history and that is being used in
    Conversation, Muc and Private tabs
    Also letting the user enter colors or other text markups
    """
    # The history is common to all MessageInput
    global_history: ClassVar[List[str]] = []

    def __init__(self) -> None:
        HistoryInput.__init__(self)
        self.key_func["KEY_UP"] = self.key_up
        self.key_func["M-A"] = self.key_up
        self.key_func["KEY_DOWN"] = self.key_down
        self.key_func["M-B"] = self.key_down
        self.key_func['^C'] = self.enter_attrib

    def enter_attrib(self) -> None:
        """
        Read one more char (c), add the corresponding char from formats_char to the text string
        """

        def cb(attr_char: str) -> None:
            if attr_char in self.text_attributes:
                char = FORMAT_CHARS[self.text_attributes.index(attr_char)]
                self.do_command(char, False)
                self.rewrite_text()

        keyboard.continuation_keys_callback = cb

    def key_enter(self) -> str | None:
        if self.history_enter():
            return None

        txt = self.get_text()
        if len(txt) != 0:
            if not self.history or self.history[0] != txt:
                # add the message to history, but avoid duplicates
                self.history.insert(0, txt)
            self.histo_pos = -1
        self.clear_text()
        return txt


class CommandInput(HistoryInput):
    """
    An input with an help message in the left, with three given callbacks:
    one when when successfully 'execute' the command and when we abort it.
    The last callback is optional and is called on any input key
    This input is used, for example, in the RosterTab when, to replace the
    HelpMessage when a command is started
    The on_input callback
    """
    global_history: ClassVar[List[str]] = []

    def __init__(self, help_message: str, on_abort: Callable[..., bool],
                 on_success: Callable[..., bool], on_input: Callable[[str], Any] | None = None) -> None:
        HistoryInput.__init__(self)
        self.on_abort = on_abort
        self.on_success = on_success
        if on_input is not None:
            self.on_input = on_input
        else:
            self.on_input = default_on_input
        self.help_message = help_message
        self.key_func['^M'] = self.success
        self.key_func['^G'] = self.abort
        self.key_func['^C'] = self.abort
        self.key_func["KEY_UP"] = self.key_up
        self.key_func["M-A"] = self.key_up
        self.key_func["KEY_DOWN"] = self.key_down
        self.key_func["M-B"] = self.key_down

    def do_command(self, key: str, reset: bool = True, raw: bool = False) -> bool:
        res = Input.do_command(self, key, reset=reset, raw=raw)
        if self.on_input is not default_on_input:
            self.on_input(self.get_text())
        return res

    def disable_history(self) -> None:
        """
        Disable the history (up/down) keys
        """
        if 'KEY_UP' in self.key_func:
            del self.key_func['KEY_UP']
        if 'KEY_DOWN' in self.key_func:
            del self.key_func['KEY_DOWN']

    @property
    def history_disabled(self) -> bool:
        return 'KEY_UP' not in self.key_func and 'KEY_DOWN' not in self.key_func

    def success(self) -> bool:
        """
        call the success callback, passing the text as argument
        """
        self.on_input = default_on_input
        if self.search:
            self.history_enter()
        res = self.on_success(self.get_text())
        return res

    def abort(self) -> bool:
        """
        Call the abort callback, passing the text as argument
        """
        self.on_input = default_on_input
        return self.on_abort(self.get_text())

    def on_delete(self) -> None:
        """
        SERIOUSLY BIG WTF.

        I can do
        self.key_func.clear()

        but not
        del self.key_func
        because that would raise an AttributeError exception. WTF.
        """
        self.on_input = default_on_input
        self.key_func.clear()

    def key_enter(self) -> Optional[str]:
        txt = self.get_text()
        if len(txt) != 0:
            if not self.history or self.history[0] != txt:
                # add the message to history, but avoid duplicates
                self.history.insert(0, txt)
            self.histo_pos = -1
        return None



"""
This plugin adds a :term:`/vcard` command to all tabs, allowing you to request
and display the vcard-temp of any given entity.

Command
-------

.. glossary::

    /vcard

        **Usage (globally):** ``/vcard <jid>``

        **Usage (in a chatroom tab):** ``/vcard <jid or nick>``

        **Usage (in a conversation or contact list tab):** ``/vcard [jid]``

        Globally, you can do ``/vcard user@server.example`` to get a vcard.

        In a chatroom , you can either do it on a JID or a nick (``/vcard nick``,
        ``/vcard room@muc.server.example/nick`` or ``/vcard
        user@server.example``).

        In a private or a direct conversation, you can do ``/vcard`` to request
        vcard from the current interlocutor, and in the contact list to do it
        on the currently selected contact.
"""

from poezio.decorators import command_args_parser
from poezio.plugin import BasePlugin
from poezio.roster import roster
from poezio.contact import Contact, Resource
from poezio.windows.inputs import Input
from poezio.core.structs import Completion
from poezio import tabs
from slixmpp import JID, InvalidJID, Iq
from slixmpp.exceptions import IqTimeout
from slixmpp.plugins.xep_0004 import Form


class Plugin(BasePlugin):
    def init(self) -> None:
        self.api.add_command(
            'vcard',
            self.command_vcard,
            usage='<jid>',
            help='Send an XMPP vcard request to jid (see XEP-0054).',
            short='Send a vcard request',
            completion=self.completion_vcard)
        self.api.add_tab_command(
            tabs.MucTab,
            'vcard',
            self.command_muc_vcard,
            usage='<jid|nick>',
            help='Send an XMPP vcard request to jid or nick (see XEP-0054).',
            short='Send a vcard request.',
            completion=self.completion_muc_vcard)
        self.api.add_tab_command(
            tabs.RosterInfoTab,
            'vcard',
            self.command_roster_vcard,
            usage='<jid>',
            help='Send an XMPP vcard request to jid (see XEP-0054).',
            short='Send a vcard request.',
            completion=self.completion_vcard)
        for _class in (tabs.PrivateTab, tabs.DynamicConversationTab, tabs.StaticConversationTab):
            self.api.add_tab_command(
                _class,
                'vcard',
                self.command_private_vcard,
                usage='[jid]',
                help=(
                    'Send an XMPP vcard request to the current '
                    'interlocutor or the given JID.'
                ),
                short='Send a vcard request',
                completion=self.completion_vcard)

    def _handle_vcard(self, iq: Iq) -> None:
        '''Retrieves a vCard from vcard-temp and present it as a DataFormsTab.
        '''
        jid = iq['from']

        if iq['type'] == 'error':
            self.api.information(
                'Error retrieving vCard for %s: %s: %s' %
                (jid, iq['error']['type'], iq['error']['condition']), 'Error')
            return

        vcard = iq['vcard_temp']

        form = self.core.xmpp['xep_0004'].make_form(
            ftype='result', title='vCard of %s' % jid)

        # TODO: implement the other fields.

        form.add_field(
            var='FN', ftype='text-single', label='Name', value=vcard['FN'])
        form.add_field(
            var='NICKNAME',
            ftype='text-multi',
            label='Nicknames',
            value=vcard['NICKNAME'])

        # TODO: find a way to detect whether this is present or not.
        form.add_field(ftype='fixed', value='Full Name')
        form.add_field(
            var='N/GIVEN',
            ftype='text-single',
            label='Given',
            value=vcard['N']['GIVEN'])
        form.add_field(
            var='N/MIDDLE',
            ftype='text-single',
            label='Middle',
            value=vcard['N']['MIDDLE'])
        form.add_field(
            var='N/FAMILY',
            ftype='text-single',
            label='Family',
            value=vcard['N']['FAMILY'])
        form.add_field(
            var='N/PREFIX',
            ftype='text-single',
            label='Prefix',
            value=vcard['N']['PREFIX'])
        form.add_field(
            var='N/SUFFIX',
            ftype='text-single',
            label='Suffix',
            value=vcard['N']['SUFFIX'])

        for i, addr in enumerate(vcard['addresses']):
            form.add_field(ftype='fixed', value='Address')
            values = [type_ for type_ in addr.bool_interfaces if addr[type_]]
            addr_type = form.add_field(
                var='ADR %d/TYPE' % i,
                ftype='list-multi',
                label='Type',
                value=values)
            addr_type.add_option(label='Home', value='HOME')
            for type_ in addr.bool_interfaces:
                addr_type.add_option(label=type_, value=type_)
            form.add_field(
                var='ADR %d/POBOX' % i,
                ftype='text-single',
                label='Pobox',
                value=addr['POBOX'])
            form.add_field(
                var='ADR %d/EXTADD' % i,
                ftype='text-single',
                label='Extended Address',
                value=addr['EXTADD'])
            form.add_field(
                var='ADR %d/STREET' % i,
                ftype='text-single',
                label='Street',
                value=addr['STREET'])
            form.add_field(
                var='ADR %d/LOCALITY' % i,
                ftype='text-single',
                label='Locality',
                value=addr['LOCALITY'])
            form.add_field(
                var='ADR %d/REGION' % i,
                ftype='text-single',
                label='Region',
                value=addr['REGION'])
            form.add_field(
                var='ADR %d/PCODE' % i,
                ftype='text-single',
                label='Post Code',
                value=addr['PCODE'])
            form.add_field(
                var='ADR %d/CTRY' % i,
                ftype='text-single',
                label='Country',
                value=addr['CTRY'])

        for i, tel in enumerate(vcard['telephone_numbers']):
            form.add_field(ftype='fixed', value='Telephone')
            values = [type_ for type_ in tel.bool_interfaces if tel[type_]]
            tel_type = form.add_field(
                var='TEL %d/TYPE' % i,
                ftype='list-multi',
                label='Type',
                value=values)
            for type_ in tel.bool_interfaces:
                tel_type.add_option(label=type_, value=type_)
            form.add_field(
                var='TEL %d/NUMBER' % i,
                ftype='text-single',
                label='Number',
                value=tel['NUMBER'])

        for i, email in enumerate(vcard['emails']):
            form.add_field(ftype='fixed', value='Email address')
            values = [type_ for type_ in email.bool_interfaces if email[type_]]
            email_type = form.add_field(
                var='EMAIL %d/TYPE' % i,
                ftype='list-multi',
                label='Type',
                value=values)
            for type_ in email.bool_interfaces:
                email_type.add_option(label=type_, value=type_)
            form.add_field(
                var='EMAIL %d/USERID' % i,
                ftype='text-single',
                label='Email Address',
                value=email['USERID'])

        form.add_field(ftype='fixed', value='Misc')
        form.add_field(
            var='BDAY',
            ftype='text-single',
            label='Birthday',
            value=str(vcard['BDAY']))

        for i, jabberid in enumerate(vcard['jids']):
            form.add_field(ftype='fixed', value='URL')
            form.add_field(
                var='JABBERID %d' % i,
                ftype='jid-single',
                label='URL',
                value=jabberid['JABBERID'])

        for i, url in enumerate(vcard['urls']):
            form.add_field(ftype='fixed', value='URL')
            form.add_field(
                var='URL %d' % i,
                ftype='text-single',
                label='URL',
                value=url['URL'])

        for i, desc in enumerate(vcard['descriptions']):
            form.add_field(ftype='fixed', value='Description')
            form.add_field(
                var='DESC %d' % i,
                ftype='text-multi',
                label='Description',
                value=desc['DESC'])

        def handler(form: Form) -> None:
            self.core.close_tab()

        on_validate = on_cancel = handler
        self.core.open_new_form(form, on_cancel, on_validate)

    async def _get_vcard(self, jid: JID | str) -> None:
        '''Send an iq to ask the vCard.'''
        real_jid = JID(jid)
        try:
            vcard = await self.core.xmpp.plugin['xep_0054'].get_vcard(
                jid=real_jid,
                timeout=30,
            )
            self._handle_vcard(vcard)
        except IqTimeout:
            self.api.information('Timeout while retrieving vCard for %s' % jid,
                                 'Error')

    @command_args_parser.raw
    def command_vcard(self, arg: str) -> None:
        if not arg:
            self.core.command.help('vcard')
            return

        try:
            jid = JID(arg)
        except InvalidJID:
            self.api.information('Invalid JID: %s' % arg, 'Error')
            return

        self.core.register_task(
            self._get_vcard(jid)
        )

    @command_args_parser.raw
    def command_private_vcard(self, arg: str) -> None:
        if arg:
            self.command_vcard(arg)
            return
        tab = self.api.current_tab()
        if not isinstance(tab, tabs.OneToOneTab):
            return
        self.command_vcard(tab.jid.full)

    @command_args_parser.raw
    def command_muc_vcard(self, arg: str) -> None:
        if not arg:
            self.core.command.help('vcard')
            return
        tab = self.api.current_tab()
        if not isinstance(tab, tabs.MucTab):
            return
        user = tab.get_user_by_name(arg)
        if user:
            jid = JID(tab.jid.bare + '/' + user.nick)
        else:
            try:
                jid = JID(arg)
            except InvalidJID:
                self.api.information('Invalid JID: %s' % arg, 'Error')
                return
        self.core.register_task(
            self._get_vcard(jid)
        )

    @command_args_parser.raw
    def command_roster_vcard(self, arg: str) -> None:
        if arg:
            self.command_vcard(arg)
            return
        tab = self.api.current_tab()
        if not isinstance(tab, tabs.RosterInfoTab):
            return
        current = tab.selected_row
        if isinstance(current, Resource):
            self.core.register_task(
                self._get_vcard(JID(current.jid).bare)
            )
        elif isinstance(current, Contact):
            self.core.register_task(
                self._get_vcard(current.bare_jid)
            )

    def completion_vcard(self, the_input: Input) -> Completion:
        contacts = [str(contact.bare_jid) for contact in roster.get_contacts()]
        return Completion(
            the_input.auto_completion, contacts, '', quotify=False)

    def completion_muc_vcard(self, the_input: Input) -> Completion | None:
        tab = self.api.current_tab()
        if not isinstance(tab, tabs.MucTab):
            return None
        users = [user.nick for user in tab.users]
        users.extend([
            str(resource.jid) for contact in roster.get_contacts()
            for resource in contact.resources
        ])
        return Completion(the_input.auto_completion, users, '', quotify=False)



"""
Once loaded, every line of your messages will be stripped of their trailing spaces.
"""
from poezio.plugin import BasePlugin
from slixmpp import Message
from typing import Any


class Plugin(BasePlugin):
    def init(self) -> None:
        self.api.add_event_handler('muc_say', self.rstrip)
        self.api.add_event_handler('conversation_say', self.rstrip)
        self.api.add_event_handler('private_say', self.rstrip)

    def rstrip(self, msg: Message, tab: Any) -> None:
        msg['body'] = '\n'.join(
            line.rstrip() for line in msg['body'].split('\n'))



"""
This modules contains a class that loads messages into a ChatTab, either from
MAM or the local logs, and a class that loads MUC history into the local
logs.


How the log loading works will depend on the poezio configuration:

- if use_log is True, no logs will be fetched dynamically
- if use_log is False, all logs will be fetched from MAM (if available)
- if mam_sync and use_log are True, most chat tabs (all of them except the
  static conversation tab) will try to sync the local
  logs with the MAM history when opening them, or when joining a room.
- all log loading/writing workflows are paused until the MAM sync is complete
  (so that the local log loading can be up-to-date with the MAM history)
- when use_log is False, mam_sync has no effect
"""
from __future__ import annotations
import asyncio
import logging
from datetime import datetime, timedelta, timezone
from typing import List, Optional, Sequence
from poezio import tabs
from poezio.logger import (
    build_log_message,
    iterate_messages_reverse,
    last_message_in_archive,
    Logger,
    LogDict,
)
from poezio.mam import (
    fetch_history,
    NoMAMSupportException,
    MAMQueryException,
    DiscoInfoException,
    make_line,
)
from poezio.common import to_utc
from poezio.ui.types import EndOfArchive, Message, BaseMessage
from poezio.text_buffer import HistoryGap
from slixmpp import JID


# Max number of messages to insert when filling a gap
HARD_LIMIT = 999


log = logging.getLogger(__name__)


def make_line_local(tab: tabs.ChatTab, msg: LogDict) -> Message:
    """Create a UI message from a local log read.

    :param tab: Tab in which that message will be displayed
    :param msg: Log data
    :returns: The UI message
    """
    if isinstance(tab, tabs.MucTab):
        jid = JID(tab.jid)
        jid.resource = msg.get('nickname') or ''
    else:
        jid = JID(tab.jid)
    msg['time'] = msg['time'].astimezone(tz=timezone.utc)
    return make_line(tab, msg['txt'], msg['time'], jid, '', msg['nickname'])


class LogLoader:
    """
    An ephemeral class that loads history in a tab.

    Loading from local logs is blocked until history has been fetched from
    MAM to fill the local archive.
    """
    __slots__ = ('logger', 'tab', 'mam_only', 'done_event')
    logger: Logger
    tab: tabs.ChatTab
    mam_only: bool
    done_event: asyncio.Event | None

    def __init__(self, logger: Logger, tab: tabs.ChatTab,
                 local_logs: bool = True,
                 done_event: Optional[asyncio.Event] = None):
        self.mam_only = not local_logs
        self.logger = logger
        self.tab = tab
        self.done_event = done_event

    def _done(self) -> None:
        """Signal end if possible"""
        if self.done_event is not None:
            self.done_event.set()

    async def tab_open(self) -> None:
        """Called on a tab opening or a MUC join"""
        amount = 2 * self.tab.text_win.height
        gap = self.tab._text_buffer.find_last_gap_muc()
        messages: Sequence[BaseMessage]
        if gap is not None:
            if self.mam_only:
                messages = await self.mam_fill_gap(gap, amount)
            else:
                messages = await self.local_fill_gap(gap, amount)
        else:
            if self.mam_only:
                messages = await self.mam_tab_open(amount)
            else:
                messages = await self.local_tab_open(amount)

        log.debug(
            'Fetched %s messages for %s',
            len(messages), self.tab.jid
        )
        if messages:
            self.tab._text_buffer.add_history_messages(messages, gap=gap)
            self.tab.core.refresh_window()
        self._done()

    async def mam_tab_open(self, nb: int) -> Sequence[BaseMessage]:
        """Fetch messages in MAM when opening a new tab.

        :param nb: number of max messages to fetch.
        :returns: list of ui messages to add
        """
        tab = self.tab
        end = datetime.now()
        for message in tab._text_buffer.messages:
            time_ok = to_utc(message.time) < to_utc(end)
            if isinstance(message, Message) and time_ok:
                end = message.time
                break
        end = end - timedelta(microseconds=1)
        try:
            return await fetch_history(tab, end=end, amount=nb)
        except (NoMAMSupportException, MAMQueryException, DiscoInfoException):
            return []
        finally:
            tab.query_status = False

    def _get_time_limit(self) -> datetime:
        """Get the date 10 weeks ago from now."""
        return datetime.now() - timedelta(weeks=10)

    async def local_tab_open(self, nb: int) -> Sequence[BaseMessage]:
        """Fetch messages locally when opening a new tab.

        :param nb: number of max messages to fetch.
        :returns: list of ui messages to add
        """
        await self.wait_mam()
        limit = self._get_time_limit()
        results: List[BaseMessage] = []
        filepath = self.logger.get_file_path(self.tab.jid)
        count = 0
        for msg in iterate_messages_reverse(filepath):
            typ_ = msg.pop('type')
            if typ_ == 'message':
                results.append(make_line_local(self.tab, msg))
            elif msg['time'] < limit and 'set the subject' not in msg['txt']:
                break
            if len(results) >= nb:
                break
            count += 1
            if count % 20 == 0:
                await asyncio.sleep(0)
        return results[::-1]

    async def mam_fill_gap(self, gap: HistoryGap,
                           amount: Optional[int] = None) -> Sequence[BaseMessage]:
        """Fill a message gap in an existing tab using MAM.

        :param gap: Object describing the history gap
        :returns: list of ui messages to add
        """
        tab = self.tab
        if amount is None:
            amount = HARD_LIMIT

        start = gap.last_timestamp_before_leave
        end = gap.first_timestamp_after_join
        if start:
            start = start + timedelta(seconds=1)
        if end:
            end = end - timedelta(seconds=1)
        try:
            return await fetch_history(
                tab,
                start=start,
                end=end,
                amount=amount,
            )
        except (NoMAMSupportException, MAMQueryException, DiscoInfoException):
            return []
        finally:
            tab.query_status = False

    async def local_fill_gap(self, gap: HistoryGap,
                             amount: Optional[int] = None) -> Sequence[BaseMessage]:
        """Fill a message gap in an existing tab using the local logs.
        Mostly useless when not used with the MAMFiller.

        :param gap: Object describing the history gap
        :returns: list of ui messages to add
        """
        if amount is None:
            amount = HARD_LIMIT
        await self.wait_mam()
        limit = self._get_time_limit()
        start = gap.last_timestamp_before_leave
        end = gap.first_timestamp_after_join
        count = 0

        results: List[BaseMessage] = []
        filepath = self.logger.get_file_path(self.tab.jid)
        for msg in iterate_messages_reverse(filepath):
            typ_ = msg.pop('type')
            if start and msg['time'] < start:
                break
            if typ_ == 'message' and (not end or msg['time'] < end):
                results.append(make_line_local(self.tab, msg))
            elif msg['time'] < limit and 'set the subject' not in msg['txt']:
                break
            if len(results) >= amount:
                break
            count += 1
            if count % 20 == 0:
                await asyncio.sleep(0)
        return results[::-1]

    async def scroll_requested(self) -> None:
        """When a scroll up is requested in a chat tab.

        Try to load more history if there are no more messages in the buffer.
        """
        tab = self.tab
        tw = tab.text_win

        # If position in the tab is < two screen pages, then fetch MAM, so that
        # wa keep some prefetched margin. A first page should also be
        # prefetched on join if not already available.
        total, pos, height = len(tw.built_lines), tw.pos, tw.height
        rest = (total - pos) // height

        if rest > 1:
            return

        messages: Sequence[BaseMessage]
        if self.mam_only:
            messages = await self.mam_scroll_requested(height)
        else:
            messages = await self.local_scroll_requested(height)
        if messages:
            tab._text_buffer.add_history_messages(messages)
            tab.core.refresh_window()
        self._done()

    async def local_scroll_requested(self, nb: int) -> Sequence[BaseMessage]:
        """Fetch messages locally on scroll up.

        :param nb: Number of messages to fetch
        :returns: list of ui messages to add
        """
        await self.wait_mam()
        tab = self.tab
        count = 0

        first_message = tab._text_buffer.find_first_message()
        first_message_time = None
        if first_message:
            first_message_time = first_message.time - timedelta(microseconds=1)

        results: List[BaseMessage] = []
        filepath = self.logger.get_file_path(self.tab.jid)
        for msg in iterate_messages_reverse(filepath):
            typ_ = msg.pop('type')
            if first_message_time is None or msg['time'] < first_message_time:
                if typ_ == 'message':
                    results.append(make_line_local(self.tab, msg))
            if len(results) >= nb:
                break
            count += 1
            if count % 20 == 0:
                await asyncio.sleep(0)
        return results[::-1]

    async def mam_scroll_requested(self, nb: int) -> Sequence[BaseMessage]:
        """Fetch messages from MAM on scroll up.

        :param nb: Number of messages to fetch
        :returns: list of ui messages to add
        """
        tab = self.tab
        try:
            messages: Sequence[BaseMessage] = await fetch_history(tab, amount=nb)
        except NoMAMSupportException:
            return []
        except (MAMQueryException, DiscoInfoException):
            tab.core.information(
                f'An error occurred when fetching MAM for {tab.jid}',
                'Error'
            )
            return []
        else:
            last_message_exists = False
            last_message = None
            if tab._text_buffer.messages:
                last_message = tab._text_buffer.messages[0]
                last_message_exists = True
            if (not messages and
                    last_message_exists
                    and not isinstance(last_message, EndOfArchive)):
                time = tab._text_buffer.messages[0].time
                messages = [EndOfArchive('End of archive reached', time=time)]
            return messages
        finally:
            tab.query_status = False

    async def wait_mam(self) -> None:
        """Wait for the MAM history sync before reading the local logs.

        Does nothing apart from blocking.
        """
        if self.tab.mam_filler is None:
            return
        await self.tab.mam_filler.done.wait()


class MAMFiller:
    """Class that loads messages from MAM history into the local logs.
    """
    __slots__ = ('tab', 'end_date', 'logger', 'future',
                 'done', 'limit', 'end_date', 'result')
    tab: tabs.ChatTab
    logger: Logger
    future: asyncio.Task[None]
    done: asyncio.Event
    limit: int
    result: int
    end_date: datetime | None

    def __init__(self, logger: Logger, tab: tabs.ChatTab, limit: int = 2000,
                 end: datetime | None = None) -> None:
        self.tab = tab
        self.end_date = end
        self.logger = logger
        logger.fd_busy(tab.jid)
        self.future = asyncio.create_task(self.fetch_routine())
        self.done = asyncio.Event()
        self.limit = limit
        self.result = 0

    def cancel(self) -> None:
        """Cancel the routine and signal the end."""
        self.future.cancel()
        self.end()

    async def fetch_routine(self) -> None:
        """Load logs into the local archive, if possible."""
        filepath = self.logger.get_file_path(self.tab.jid)
        log.debug('Fetching logs for %s', self.tab.jid)
        try:
            last_msg = last_message_in_archive(filepath)
            last_msg_time = None
            if last_msg:
                last_msg_time = last_msg['time'] + timedelta(seconds=1)
            try:
                messages = await fetch_history(
                    self.tab,
                    start=last_msg_time,
                    amount=self.limit,
                    end=self.end_date,
                )
                log.debug(
                    'Fetched %s messages to fill local logs for %s',
                    len(messages), self.tab.jid,
                )
                self.result = len(messages)
            except NoMAMSupportException:
                log.debug('The entity %s does not support MAM', self.tab.jid)
                return
            except (DiscoInfoException, MAMQueryException):
                log.debug(
                    'Failed fetching logs for %s',
                    self.tab.jid, exc_info=True
                )
                return

            def build_message(msg: Message) -> str:
                return build_log_message(
                    msg.nickname,
                    msg.txt,
                    msg.time,
                    prefix='MR',
                )

            logs = ''.join(map(build_message, messages))
            self.logger.log_raw(self.tab.jid, logs, force=True)
        finally:
            self.end()

    def end(self) -> None:
        """End a MAM fill (error or success). Remove references and signal on
        the Event().
        """
        try:
            self.logger.fd_available(self.tab.jid)
        except Exception:
            log.exception('Error when restoring log fd:')
        self.tab.mam_filler = None
        self.done.set()
